#!/usr/bin/env python3
#
# vmtools.py - part of the FDroid server tools
# Copyright (C) 2017 Michael Poehn <michael.poehn@fsfe.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from os.path import isdir, isfile, basename, abspath, expanduser
import os
import json
import shutil
import subprocess
import textwrap
import logging
from .common import FDroidException

import threading

lock = threading.Lock()


def get_clean_builder(serverdir):
    if not os.path.isdir(serverdir):
        if os.path.islink(serverdir):
            os.unlink(serverdir)
        logging.info("buildserver path does not exists, creating %s", serverdir)
        os.makedirs(serverdir)
    vagrantfile = os.path.join(serverdir, 'Vagrantfile')
    if not os.path.isfile(vagrantfile):
        with open(vagrantfile, 'w') as f:
            f.write(
                textwrap.dedent(
                    """\
                # generated file, do not change.

                Vagrant.configure("2") do |config|
                    config.vm.box = "buildserver"
                    config.vm.synced_folder ".", "/vagrant", disabled: true
                end
                """
                )
            )
    vm = get_build_vm(serverdir)
    logging.info('destroying buildserver before build')
    vm.destroy()
    logging.info('starting buildserver')
    vm.up()

    try:
        sshinfo = vm.sshinfo()
    except FDroidBuildVmException:
        # workaround because libvirt sometimes likes to forget
        # about ssh connection info even thou the vm is running
        vm.halt()
        vm.up()
        sshinfo = vm.sshinfo()

    return sshinfo


def _check_call(cmd, cwd=None):
    logging.debug(' '.join(cmd))
    return subprocess.check_call(cmd, shell=False, cwd=cwd)


def _check_output(cmd, cwd=None):
    logging.debug(' '.join(cmd))
    return subprocess.check_output(cmd, shell=False, cwd=cwd)


def get_build_vm(srvdir, provider=None):
    """No summary.

    Factory function for getting FDroidBuildVm instances.

    This function tries to figure out what hypervisor should be used
    and creates an object for controlling a build VM.

    Parameters
    ----------
    srvdir
      path to a directory which contains a Vagrantfile
    provider
      optionally this parameter allows specifiying an
      specific vagrant provider.

    Returns
    -------
    FDroidBuildVm instance.
    """
    abssrvdir = abspath(srvdir)

    # use supplied provider
    if provider:
        if provider == 'libvirt':
            logging.debug('build vm provider \'libvirt\' selected')
            return LibvirtBuildVm(abssrvdir)
        elif provider == 'virtualbox':
            logging.debug('build vm provider \'virtualbox\' selected')
            return VirtualboxBuildVm(abssrvdir)
        else:
            logging.warning('build vm provider not supported: \'%s\'', provider)

    # try guessing provider from installed software
    kvm_installed = shutil.which('kvm') is not None
    kvm_installed |= shutil.which('qemu') is not None
    kvm_installed |= shutil.which('qemu-kvm') is not None
    vbox_installed = shutil.which('VBoxHeadless') is not None
    if kvm_installed and vbox_installed:
        logging.debug('both kvm and vbox are installed.')
    elif kvm_installed:
        logging.debug(
            'libvirt is the sole installed and supported vagrant provider, selecting \'libvirt\''
        )
        return LibvirtBuildVm(abssrvdir)
    elif vbox_installed:
        logging.debug(
            'virtualbox is the sole installed and supported vagrant provider, selecting \'virtualbox\''
        )
        return VirtualboxBuildVm(abssrvdir)
    else:
        logging.debug(
            'could not confirm that either virtualbox or kvm/libvirt are installed'
        )

    # try guessing provider from .../srvdir/.vagrant internals
    vagrant_libvirt_path = os.path.join(
        abssrvdir, '.vagrant', 'machines', 'default', 'libvirt'
    )
    has_libvirt_machine = (
        isdir(vagrant_libvirt_path) and len(os.listdir(vagrant_libvirt_path)) > 0
    )
    vagrant_virtualbox_path = os.path.join(
        abssrvdir, '.vagrant', 'machines', 'default', 'virtualbox'
    )
    has_vbox_machine = (
        isdir(vagrant_virtualbox_path) and len(os.listdir(vagrant_virtualbox_path)) > 0
    )
    if has_libvirt_machine and has_vbox_machine:
        logging.info(
            'build vm provider lookup found virtualbox and libvirt, defaulting to \'virtualbox\''
        )
        return VirtualboxBuildVm(abssrvdir)
    elif has_libvirt_machine:
        logging.debug('build vm provider lookup found \'libvirt\'')
        return LibvirtBuildVm(abssrvdir)
    elif has_vbox_machine:
        logging.debug('build vm provider lookup found \'virtualbox\'')
        return VirtualboxBuildVm(abssrvdir)

    # try guessing provider from available buildserver boxes
    available_boxes = []
    import vagrant

    boxes = vagrant.Vagrant().box_list()
    for box in boxes:
        if box.name == "buildserver":
            available_boxes.append(box.provider)
    if "libvirt" in available_boxes and "virtualbox" in available_boxes:
        logging.info(
            'basebox lookup found virtualbox and libvirt boxes, defaulting to \'virtualbox\''
        )
        return VirtualboxBuildVm(abssrvdir)
    elif "libvirt" in available_boxes:
        logging.info('\'libvirt\' buildserver box available, using that')
        return LibvirtBuildVm(abssrvdir)
    elif "virtualbox" in available_boxes:
        logging.info('\'virtualbox\' buildserver box available, using that')
        return VirtualboxBuildVm(abssrvdir)
    else:
        logging.error('No available \'buildserver\' box. Cannot proceed')
        os._exit(1)


class FDroidBuildVmException(FDroidException):
    pass


class FDroidBuildVm:
    """Abstract base class for working with FDroids build-servers.

    Use the factory method `fdroidserver.vmtools.get_build_vm()` for
    getting correct instances of this class.

    This is intended to be a hypervisor independent, fault tolerant
    wrapper around the vagrant functions we use.
    """

    def __init__(self, srvdir):
        """Create new server class."""
        self.srvdir = srvdir
        self.srvname = basename(srvdir) + '_default'
        self.vgrntfile = os.path.join(srvdir, 'Vagrantfile')
        self.srvuuid = self._vagrant_fetch_uuid()
        if not isdir(srvdir):
            raise FDroidBuildVmException(
                "Can not init vagrant, directory %s not present" % (srvdir)
            )
        if not isfile(self.vgrntfile):
            raise FDroidBuildVmException(
                "Can not init vagrant, '%s' not present" % (self.vgrntfile)
            )
        import vagrant

        self.vgrnt = vagrant.Vagrant(
            root=srvdir, out_cm=vagrant.stdout_cm, err_cm=vagrant.stdout_cm
        )

    def up(self, provision=True):
        global lock
        with lock:
            try:
                self.vgrnt.up(provision=provision, provider=self.provider)
                self.srvuuid = self._vagrant_fetch_uuid()
            except subprocess.CalledProcessError as e:
                statusline = ""
                try:
                    # try to get some additional info about the vagrant vm
                    status = self.vgrnt.status()
                    if len(status) > 0:
                        statusline = "VM status: name={n}, state={s}, provider={p}"\
                            .format(n=status[0].name,
                                    s=status[0].state,
                                    p=status[0].provider)
                except subprocess.CalledProcessError:
                    pass
                raise FDroidBuildVmException(value="could not bring up vm '{vmname}'"
                                                   .format(vmname=self.srvname),
                                             detail="{err}\n{statline}"
                                                    .format(err=str(e), statline=statusline)
                                             ) from e

    def suspend(self):
        global lock
        with lock:
            logging.info('suspending buildserver')
            try:
                self.vgrnt.suspend()
            except subprocess.CalledProcessError as e:
                raise FDroidBuildVmException("could not suspend vm '%s'" % self.srvname) from e

    def halt(self):
        global lock
        with lock:
            self.vgrnt.halt(force=True)

    def destroy(self):
        """Remove every trace of this VM from the system.

        This includes deleting:
        * hypervisor specific definitions
        * vagrant state informations (eg. `.vagrant` folder)
        * images related to this vm
        """
        logging.info("destroying vm '%s'", self.srvname)
        try:
            self.vgrnt.destroy()
            logging.debug('vagrant destroy completed')
        except subprocess.CalledProcessError as e:
            logging.exception('vagrant destroy failed: %s', e)
        vgrntdir = os.path.join(self.srvdir, '.vagrant')
        try:
            shutil.rmtree(vgrntdir)
            logging.debug('deleted vagrant dir: %s', vgrntdir)
        except Exception as e:
            logging.debug("could not delete vagrant dir: %s, %s", vgrntdir, e)
        try:
            _check_call(['vagrant', 'global-status', '--prune'])
        except subprocess.CalledProcessError as e:
            logging.debug('pruning global vagrant status failed: %s', e)

    def vagrant_uuid_okay(self):
        """Having an uuid means that vagrant up has run successfully."""
        if self.srvuuid is None:
            return False
        return True

    def _vagrant_file_name(self, name):
        return name.replace('/', '-VAGRANTSLASH-')

    def _vagrant_fetch_uuid(self):
        if isfile(os.path.join(self.srvdir, '.vagrant')):
            # Vagrant 1.0 - it's a json file...
            with open(os.path.join(self.srvdir, '.vagrant')) as f:
                id = json.load(f)['active']['default']
                logging.debug('vm uuid: %s', id)
            return id
        elif isfile(os.path.join(self.srvdir, '.vagrant', 'machines',
                                 'default', self.provider, 'id')):
            # Vagrant 1.2 (and maybe 1.1?) it's a directory tree...
            with open(os.path.join(self.srvdir, '.vagrant', 'machines',
                                   'default', self.provider, 'id')) as f:
                id = f.read()
                logging.debug('vm uuid: %s', id)
            return id
        else:
            logging.debug('vm uuid is None')
            return None

    def box_add(self, boxname, boxfile, force=True):
        """Add vagrant box to vagrant.

        Parameters
        ----------
        boxname
          name assigned to local deployment of box
        boxfile
          path to box file
        force
          overwrite existing box image (default: True)
        """
        boxfile = abspath(boxfile)
        if not isfile(boxfile):
            raise FDroidBuildVmException(
                'supplied boxfile \'%s\' does not exist' % boxfile
            )
        self.vgrnt.box_add(boxname, abspath(boxfile), force=force)

    def box_remove(self, boxname):
        try:
            _check_call(['vagrant', 'box', 'remove', '--all', '--force', boxname])
        except subprocess.CalledProcessError as e:
            logging.debug('tried removing box %s, but is did not exist: %s', boxname, e)
        boxpath = os.path.join(
            expanduser('~'), '.vagrant', self._vagrant_file_name(boxname)
        )
        if isdir(boxpath):
            logging.info(
                "attempting to remove box '%s' by deleting: %s", boxname, boxpath
            )
            shutil.rmtree(boxpath)

    def sshinfo(self):
        """Get ssh connection info for a vagrant VM.

        Returns
        -------
        A dictionary containing 'hostname', 'port', 'user' and 'idfile'
        """
        import paramiko

        try:
            sshconfig_path = os.path.join(self.srvdir, 'sshconfig')
            with open(sshconfig_path, 'wb') as fp:
                fp.write(_check_output(['vagrant', 'ssh-config'], cwd=self.srvdir))
            vagranthost = 'default'  # Host in ssh config file
            sshconfig = paramiko.SSHConfig()
            with open(sshconfig_path, 'r') as f:
                sshconfig.parse(f)
            sshconfig = sshconfig.lookup(vagranthost)
            idfile = sshconfig['identityfile']
            if isinstance(idfile, list):
                idfile = idfile[0]
            elif idfile.startswith('"') and idfile.endswith('"'):
                idfile = idfile[1:-1]
            return {
                'hostname': sshconfig['hostname'],
                'port': int(sshconfig['port']),
                'user': sshconfig['user'],
                'idfile': idfile,
            }
        except subprocess.CalledProcessError as e:
            raise FDroidBuildVmException("Error getting ssh config") from e


class LibvirtBuildVm(FDroidBuildVm):
    def __init__(self, srvdir):
        self.provider = 'libvirt'
        super().__init__(srvdir)
        import libvirt

        try:
            self.conn = libvirt.open('qemu:///system')
        except libvirt.libvirtError as e:
            raise FDroidBuildVmException('could not connect to libvirtd: %s' % (e)) from e

    def destroy(self):

        super().destroy()

        # resorting to virsh instead of libvirt python bindings, because
        # this is way more easy and therefore fault tolerant.
        # (eg. lookupByName only works on running VMs)
        try:
            _check_call(('virsh', '-c', 'qemu:///system', 'destroy', self.srvname))
        except subprocess.CalledProcessError as e:
            logging.info("could not force libvirt domain '%s' off: %s", self.srvname, e)
        try:
            # libvirt python bindings do not support all flags required
            # for undefining domains correctly.
            _check_call(('virsh', '-c', 'qemu:///system', 'undefine', self.srvname, '--nvram', '--managed-save', '--remove-all-storage', '--snapshots-metadata'))
        except subprocess.CalledProcessError as e:
            logging.info("could not undefine libvirt domain '%s': %s", self.srvname, e)

    def box_add(self, boxname, boxfile, force=True):
        boximg = '%s_vagrant_box_image_0.img' % (boxname)
        if force:
            try:
                _check_call(['virsh', '-c', 'qemu:///system', 'vol-delete', '--pool', 'default', boximg])
                logging.debug("removed old box image '%s'"
                              "from libvirt storeage pool", boximg)
            except subprocess.CalledProcessError as e:
                logging.debug("tried removing old box image '%s',"
                              "file was not present in first place",
                              boximg, exc_info=e)
        super().box_add(boxname, boxfile, force)

    def box_remove(self, boxname):
        super().box_remove(boxname)
        try:
            _check_call(['virsh', '-c', 'qemu:///system', 'vol-delete', '--pool', 'default', '%s_vagrant_box_image_0.img' % (boxname)])
        except subprocess.CalledProcessError as e:
            logging.debug("tried removing '%s', file was not present in first place", boxname, exc_info=e)


class VirtualboxBuildVm(FDroidBuildVm):

    def __init__(self, srvdir):
        self.provider = 'virtualbox'
        super().__init__(srvdir)



#!/usr/bin/env python3
#
# an fdroid plugin for exporting a repo's keystore in standard PEM format

import os
from argparse import ArgumentParser
from fdroidserver import common
from fdroidserver.common import FDroidPopen
from fdroidserver.exception import BuildException

fdroid_summary = "export the repo's keystore file to a NitroKey HSM"


def run(cmd, error):
    envs = {'LC_ALL': 'C.UTF-8',
            'PIN': config['smartcard_pin'],
            'FDROID_KEY_STORE_PASS': config['keystorepass'],
            'FDROID_KEY_PASS': config['keypass']}
    p = FDroidPopen(cmd, envs=envs)
    if p.returncode != 0:
        raise BuildException(error, p.output)


def main():
    global config
    parser = ArgumentParser()
    common.setup_global_opts(parser)
    options = parser.parse_args()
    config = common.read_config(options)
    destkeystore = config['keystore'].replace('.jks', '.p12').replace('/', '_')
    exportkeystore = config['keystore'].replace('.jks', '.pem').replace('/', '_')
    if os.path.exists(destkeystore) or os.path.exists(exportkeystore):
        raise BuildException('%s exists!' % exportkeystore)
    run([config['keytool'], '-importkeystore',
         '-srckeystore', config['keystore'],
         '-srcalias', config['repo_keyalias'],
         '-srcstorepass:env', 'FDROID_KEY_STORE_PASS',
         '-srckeypass:env', 'FDROID_KEY_PASS',
         '-destkeystore', destkeystore,
         '-deststorepass:env', 'FDROID_KEY_STORE_PASS',
         '-deststoretype', 'PKCS12'],
        'Failed to convert to PKCS12!')
#    run(['openssl', 'pkcs12', '-in', destkeystore,
#         '-passin', 'env:FDROID_KEY_STORE_PASS', '-nokeys',
#         '-out', exportkeystore,
#         '-passout', 'env:FDROID_KEY_STORE_PASS'],
#        'Failed to convert to PEM!')
    run(['pkcs15-init', '--delete-objects', 'privkey,pubkey',
         '--id', '3', '--store-private-key', destkeystore,
         '--format', 'pkcs12', '--auth-id', '3',
         '--verify-pin', '--pin', 'env:PIN'],
        '')
    run(['pkcs15-init', '--delete-objects', 'privkey,pubkey',
         '--id', '2', '--store-private-key', destkeystore,
         '--format', 'pkcs12', '--auth-id', '3',
         '--verify-pin', '--pin', 'env:PIN'],
        '')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# an fdroid plugin print the repo_pubkey from a repo's keystore
#

from argparse import ArgumentParser
from fdroidserver import common, index

fdroid_summary = 'export the keystore in standard PEM format'


def main():
    parser = ArgumentParser()
    common.setup_global_opts(parser)
    options = parser.parse_args()
    common.config = common.read_config(options)
    pubkey, repo_pubkey_fingerprint = index.extract_pubkey()
    print('repo_pubkey = "%s"' % pubkey.decode())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# metadata.py - part of the FDroid server tools
# Copyright (C) 2013, Ciaran Gultnieks, ciaran@ciarang.com
# Copyright (C) 2013-2014 Daniel Martí <mvdan@mvdan.cc>
# Copyright (C) 2017-2018 Michael Pöhn <michael.poehn@fsfe.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import git
from pathlib import Path
import platform
import re
import logging
import importlib
from collections import OrderedDict

from ruamel.yaml import YAML, YAMLError

from . import common
from . import _
from .exception import MetaDataException, FDroidException

srclibs = None
warnings_action = None

# validates usernames based on a loose collection of rules from GitHub, GitLab,
# Liberapay and issuehunt.  This is mostly to block abuse.
VALID_USERNAME_REGEX = re.compile(r'^[a-z\d](?:[a-z\d/._-]){0,38}$', re.IGNORECASE)


def _warn_or_exception(value, cause=None):
    """Output warning or Exception depending on -W."""
    if warnings_action == 'ignore':
        pass
    elif warnings_action == 'error':
        if cause:
            raise MetaDataException(value) from cause
        else:
            raise MetaDataException(value)
    else:
        logging.warning(value)


yaml_app_field_order = [
    'Disabled',
    'AntiFeatures',
    'Categories',
    'License',
    'AuthorName',
    'AuthorEmail',
    'AuthorWebSite',
    'WebSite',
    'SourceCode',
    'IssueTracker',
    'Translation',
    'Changelog',
    'Donate',
    'FlattrID',
    'Liberapay',
    'OpenCollective',
    'Bitcoin',
    'Litecoin',
    '\n',
    'Name',
    'AutoName',
    'Summary',
    'Description',
    '\n',
    'RequiresRoot',
    '\n',
    'RepoType',
    'Repo',
    'Binaries',
    '\n',
    'Builds',
    '\n',
    'AllowedAPKSigningKeys',
    '\n',
    'MaintainerNotes',
    '\n',
    'ArchivePolicy',
    'AutoUpdateMode',
    'UpdateCheckMode',
    'UpdateCheckIgnore',
    'VercodeOperation',
    'UpdateCheckName',
    'UpdateCheckData',
    'CurrentVersion',
    'CurrentVersionCode',
    '\n',
    'NoSourceSince',
]


yaml_app_fields = [x for x in yaml_app_field_order if x != '\n']


class App(dict):

    def __init__(self, copydict=None):
        if copydict:
            super().__init__(copydict)
            return
        super().__init__()

        self.Disabled = None
        self.AntiFeatures = []
        self.Provides = None
        self.Categories = []
        self.License = 'Unknown'
        self.AuthorName = None
        self.AuthorEmail = None
        self.AuthorWebSite = None
        self.WebSite = ''
        self.SourceCode = ''
        self.IssueTracker = ''
        self.Translation = ''
        self.Changelog = ''
        self.Donate = None
        self.FlattrID = None
        self.Liberapay = None
        self.OpenCollective = None
        self.Bitcoin = None
        self.Litecoin = None
        self.Name = None
        self.AutoName = ''
        self.Summary = ''
        self.Description = ''
        self.RequiresRoot = False
        self.RepoType = ''
        self.Repo = ''
        self.Binaries = None
        self.AllowedAPKSigningKeys = []
        self.MaintainerNotes = ''
        self.ArchivePolicy = None
        self.AutoUpdateMode = 'None'
        self.UpdateCheckMode = 'None'
        self.UpdateCheckIgnore = None
        self.VercodeOperation = []
        self.UpdateCheckName = None
        self.UpdateCheckData = None
        self.CurrentVersion = ''
        self.CurrentVersionCode = None
        self.NoSourceSince = ''

        self.id = None
        self.metadatapath = None
        self.Builds = []
        self.comments = {}
        self.added = None
        self.lastUpdated = None

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def get_last_build(self):
        if len(self.Builds) > 0:
            return self.Builds[-1]
        else:
            return Build()


TYPE_STRING = 2
TYPE_BOOL = 3
TYPE_LIST = 4
TYPE_SCRIPT = 5
TYPE_MULTILINE = 6
TYPE_BUILD = 7
TYPE_INT = 8

fieldtypes = {
    'Description': TYPE_MULTILINE,
    'MaintainerNotes': TYPE_MULTILINE,
    'Categories': TYPE_LIST,
    'AntiFeatures': TYPE_LIST,
    'AllowedAPKSigningKeys': TYPE_LIST,
    'Builds': TYPE_BUILD,
    'VercodeOperation': TYPE_LIST,
    'CurrentVersionCode': TYPE_INT,
}


def fieldtype(name):
    name = name.replace(' ', '')
    if name in fieldtypes:
        return fieldtypes[name]
    return TYPE_STRING


# In the order in which they are laid out on files
build_flags = [
    'versionName',
    'versionCode',
    'disable',
    'commit',
    'timeout',
    'subdir',
    'submodules',
    'sudo',
    'init',
    'patch',
    'gradle',
    'maven',
    'output',
    'binary',
    'srclibs',
    'oldsdkloc',
    'encoding',
    'forceversion',
    'forcevercode',
    'rm',
    'extlibs',
    'prebuild',
    'androidupdate',
    'target',
    'scanignore',
    'scandelete',
    'build',
    'buildjni',
    'ndk',
    'preassemble',
    'gradleprops',
    'antcommands',
    'postbuild',
    'novcheck',
    'antifeatures',
]


class Build(dict):

    def __init__(self, copydict=None):
        super().__init__()
        self.disable = ''
        self.commit = None
        self.timeout = None
        self.subdir = None
        self.submodules = False
        self.sudo = ''
        self.init = ''
        self.patch = []
        self.gradle = []
        self.maven = False
        self.output = None
        self.binary = None
        self.srclibs = []
        self.oldsdkloc = False
        self.encoding = None
        self.forceversion = False
        self.forcevercode = False
        self.rm = []
        self.extlibs = []
        self.prebuild = ''
        self.androidupdate = []
        self.target = None
        self.scanignore = []
        self.scandelete = []
        self.build = ''
        self.buildjni = []
        self.ndk = None
        self.preassemble = []
        self.gradleprops = []
        self.antcommands = []
        self.postbuild = ''
        self.novcheck = False
        self.antifeatures = []
        if copydict:
            super().__init__(copydict)
            return

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def build_method(self):
        for f in ['maven', 'gradle']:
            if self.get(f):
                return f
        if self.output:
            return 'raw'
        return 'ant'

    # like build_method, but prioritize output=
    def output_method(self):
        if self.output:
            return 'raw'
        for f in ['maven', 'gradle']:
            if self.get(f):
                return f
        return 'ant'

    def ndk_path(self):
        """Return the path to the first configured NDK or an empty string."""
        ndk = self.ndk
        if isinstance(ndk, list):
            ndk = self.ndk[0]
        return common.config['ndk_paths'].get(ndk, '')


flagtypes = {
    'versionCode': TYPE_INT,
    'extlibs': TYPE_LIST,
    'srclibs': TYPE_LIST,
    'patch': TYPE_LIST,
    'rm': TYPE_LIST,
    'buildjni': TYPE_LIST,
    'preassemble': TYPE_LIST,
    'androidupdate': TYPE_LIST,
    'scanignore': TYPE_LIST,
    'scandelete': TYPE_LIST,
    'gradle': TYPE_LIST,
    'antcommands': TYPE_LIST,
    'gradleprops': TYPE_LIST,
    'sudo': TYPE_SCRIPT,
    'init': TYPE_SCRIPT,
    'prebuild': TYPE_SCRIPT,
    'build': TYPE_SCRIPT,
    'postbuild': TYPE_SCRIPT,
    'submodules': TYPE_BOOL,
    'oldsdkloc': TYPE_BOOL,
    'forceversion': TYPE_BOOL,
    'forcevercode': TYPE_BOOL,
    'novcheck': TYPE_BOOL,
    'antifeatures': TYPE_LIST,
    'timeout': TYPE_INT,
}


def flagtype(name):
    if name in flagtypes:
        return flagtypes[name]
    return TYPE_STRING


class FieldValidator():
    """Designate App metadata field types and checks that it matches.

    'name'     - The long name of the field type
    'matching' - List of possible values or regex expression
    'sep'      - Separator to use if value may be a list
    'fields'   - Metadata fields (Field:Value) of this type
    """

    def __init__(self, name, matching, fields):
        self.name = name
        self.matching = matching
        self.compiled = re.compile(matching)
        self.fields = fields

    def check(self, v, appid):
        if not v:
            return
        if type(v) == list:
            values = v
        else:
            values = [v]
        for v in values:
            if not self.compiled.match(v):
                _warn_or_exception(_("'{value}' is not a valid {field} in {appid}. Regex pattern: {pattern}")
                                   .format(value=v, field=self.name, appid=appid, pattern=self.matching))


# Generic value types
valuetypes = {
    FieldValidator("Flattr ID",
                   r'^[0-9a-z]+$',
                   ['FlattrID']),

    FieldValidator("Liberapay",
                   VALID_USERNAME_REGEX,
                   ['Liberapay']),

    FieldValidator("Open Collective",
                   VALID_USERNAME_REGEX,
                   ['OpenCollective']),

    FieldValidator("HTTP link",
                   r'^http[s]?://',
                   ["WebSite", "SourceCode", "IssueTracker", "Translation", "Changelog", "Donate"]),

    FieldValidator("Email",
                   r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$',
                   ["AuthorEmail"]),

    FieldValidator("Bitcoin address",
                   r'^(bc1|[13])[a-zA-HJ-NP-Z0-9]{25,39}$',
                   ["Bitcoin"]),

    FieldValidator("Litecoin address",
                   r'^[LM3][a-km-zA-HJ-NP-Z1-9]{26,33}$',
                   ["Litecoin"]),

    FieldValidator("Repo Type",
                   r'^(git|git-svn|svn|hg|bzr|srclib)$',
                   ["RepoType"]),

    FieldValidator("Binaries",
                   r'^http[s]?://',
                   ["Binaries"]),

    FieldValidator("AllowedAPKSigningKeys",
                   r'^[a-fA-F0-9]{64}$',
                   ["AllowedAPKSigningKeys"]),

    FieldValidator("Archive Policy",
                   r'^[0-9]+ versions$',
                   ["ArchivePolicy"]),

    FieldValidator("Anti-Feature",
                   r'^(Ads|Tracking|NonFreeNet|NonFreeDep|NonFreeAdd|UpstreamNonFree|NonFreeAssets|KnownVuln|ApplicationDebuggable|NoSourceSince|NSFW)$',
                   ["AntiFeatures"]),

    FieldValidator("Auto Update Mode",
                   r"^(Version.*|None)$",
                   ["AutoUpdateMode"]),

    FieldValidator("Update Check Mode",
                   r"^(Tags|Tags .+|RepoManifest|RepoManifest/.+|RepoTrunk|HTTP|Static|None)$",
                   ["UpdateCheckMode"])
}


# Check an app's metadata information for integrity errors
def check_metadata(app):
    for v in valuetypes:
        for k in v.fields:
            v.check(app[k], app.id)


def parse_yaml_srclib(metadatapath):

    thisinfo = {'RepoType': '',
                'Repo': '',
                'Subdir': None,
                'Prepare': None}

    if not metadatapath.exists():
        _warn_or_exception(_("Invalid scrlib metadata: '{file}' "
                             "does not exist"
                             .format(file=metadatapath)))
        return thisinfo

    with metadatapath.open("r", encoding="utf-8") as f:
        try:
            yaml = YAML(typ='safe')
            data = yaml.load(f)
            if type(data) is not dict:
                if platform.system() == 'Windows':
                    # Handle symlink on Windows
                    symlink = metadatapath.parent / metadatapath.read_text(encoding='utf-8')
                    if symlink.is_file():
                        with symlink.open("r", encoding="utf-8") as s:
                            data = yaml.load(s)
            if type(data) is not dict:
                raise YAMLError(_('{file} is blank or corrupt!')
                                .format(file=metadatapath))
        except YAMLError as e:
            _warn_or_exception(_("Invalid srclib metadata: could not "
                                 "parse '{file}'")
                               .format(file=metadatapath) + '\n'
                               + common.run_yamllint(metadatapath, indent=4),
                               cause=e)
            return thisinfo

    for key in data.keys():
        if key not in thisinfo.keys():
            _warn_or_exception(_("Invalid srclib metadata: unknown key "
                                 "'{key}' in '{file}'")
                               .format(key=key, file=metadatapath))
            return thisinfo
        else:
            if key == 'Subdir':
                if isinstance(data[key], str):
                    thisinfo[key] = data[key].split(',')
                elif isinstance(data[key], list):
                    thisinfo[key] = data[key]
                elif data[key] is None:
                    thisinfo[key] = ['']
            elif key == 'Prepare' or flagtype(key) == TYPE_SCRIPT:
                if isinstance(data[key], list):
                    thisinfo[key] = data[key]
                else:
                    thisinfo[key] = [data[key]] if data[key] else []
            else:
                thisinfo[key] = str(data[key] or '')

    return thisinfo


def read_srclibs():
    """Read all srclib metadata.

    The information read will be accessible as metadata.srclibs, which is a
    dictionary, keyed on srclib name, with the values each being a dictionary
    in the same format as that returned by the parse_yaml_srclib function.

    A MetaDataException is raised if there are any problems with the srclib
    metadata.
    """
    global srclibs

    # They were already loaded
    if srclibs is not None:
        return

    srclibs = {}

    srcdir = Path('srclibs')
    srcdir.mkdir(exist_ok=True)

    for metadatapath in sorted(srcdir.glob('*.yml')):
        srclibs[metadatapath.stem] = parse_yaml_srclib(metadatapath)


def read_metadata(appids={}, sort_by_time=False):
    """Return a list of App instances sorted newest first.

    This reads all of the metadata files in a 'data' repository, then
    builds a list of App instances from those files.  The list is
    sorted based on creation time, newest first.  Most of the time,
    the newer files are the most interesting.

    appids is a dict with appids a keys and versionCodes as values.

    """
    # Always read the srclibs before the apps, since they can use a srlib as
    # their source repository.
    read_srclibs()

    apps = OrderedDict()

    for basedir in ('metadata', 'tmp'):
        Path(basedir).mkdir(exist_ok=True)

    if appids:
        vercodes = common.read_pkg_args(appids)
        metadatafiles = common.get_metadata_files(vercodes)
    else:
        metadatafiles = list(Path('metadata').glob('*.yml')) + list(
            Path('.').glob('.fdroid.yml'))

    if sort_by_time:
        entries = ((path.stat().st_mtime, path) for path in metadatafiles)
        metadatafiles = []
        for _ignored, path in sorted(entries, reverse=True):
            metadatafiles.append(path)
    else:
        # most things want the index alpha sorted for stability
        metadatafiles = sorted(metadatafiles)

    for metadatapath in metadatafiles:
        appid = metadatapath.stem
        if appid != '.fdroid' and not common.is_valid_package_name(appid):
            _warn_or_exception(_("{appid} from {path} is not a valid Java Package Name!")
                               .format(appid=appid, path=metadatapath))
        if appid in apps:
            _warn_or_exception(_("Found multiple metadata files for {appid}")
                               .format(appid=appid))
        app = parse_metadata(metadatapath)
        check_metadata(app)
        apps[app.id] = app

    return apps


# Port legacy ';' separators
list_sep = re.compile(r'[,;]')


def split_list_values(s):
    res = []
    for v in re.split(list_sep, s):
        if not v:
            continue
        v = v.strip()
        if not v:
            continue
        res.append(v)
    return res


def sorted_builds(builds):
    return sorted(builds, key=lambda build: build.versionCode)


esc_newlines = re.compile(r'\\( |\n)')


def post_metadata_parse(app):
    if 'flavours' in app and app['flavours'] == [True]:
        app['flavours'] = 'yes'

    for k, v in app.items():
        if fieldtype(k) == TYPE_LIST:
            if isinstance(v, str):
                app[k] = [v, ]
            elif v:
                app[k] = [str(i) for i in v]
        elif fieldtype(k) == TYPE_INT:
            if v:
                app[k] = int(v)
        elif fieldtype(k) == TYPE_STRING:
            if v:
                app[k] = str(v)
        else:
            if type(v) in (float, int):
                app[k] = str(v)

    def _yaml_bool_unmapable(v):
        return v in (True, False, [True], [False])

    def _yaml_bool_unmap(v):
        if v is True:
            return 'yes'
        elif v is False:
            return 'no'
        elif v == [True]:
            return ['yes']
        elif v == [False]:
            return ['no']

    _bool_allowed = ('maven')

    builds = []
    if 'Builds' in app:
        for build in app.get('Builds', []):
            if not isinstance(build, Build):
                build = Build(build)
            for k, v in build.items():
                if not (v is None):
                    if flagtype(k) == TYPE_LIST:
                        if _yaml_bool_unmapable(v):
                            build[k] = _yaml_bool_unmap(v)

                        if isinstance(v, str):
                            build[k] = [v]
                        elif isinstance(v, bool):
                            if v:
                                build[k] = ['yes']
                            else:
                                build[k] = []
                    elif flagtype(k) is TYPE_INT:
                        build[k] = v
                    elif flagtype(k) is TYPE_STRING:
                        if isinstance(v, bool) and k in _bool_allowed:
                            build[k] = v
                        else:
                            if _yaml_bool_unmapable(v):
                                build[k] = _yaml_bool_unmap(v)
                            else:
                                build[k] = str(v)
            builds.append(build)

    app['Builds'] = sorted_builds(builds)


# Parse metadata for a single application.
#
#  'metadatapath' - the file path to read. The "Application ID" aka
#               "Package Name" for the application comes from this
#               filename. Pass None to get a blank entry.
#
# Returns a dictionary containing all the details of the application. There are
# two major kinds of information in the dictionary. Keys beginning with capital
# letters correspond directory to identically named keys in the metadata file.
# Keys beginning with lower case letters are generated in one way or another,
# and are not found verbatim in the metadata.
#
# Known keys not originating from the metadata are:
#
#  'comments'         - a list of comments from the metadata file. Each is
#                       a list of the form [field, comment] where field is
#                       the name of the field it preceded in the metadata
#                       file. Where field is None, the comment goes at the
#                       end of the file. Alternatively, 'build:version' is
#                       for a comment before a particular build version.
#  'descriptionlines' - original lines of description as formatted in the
#                       metadata file.
#


bool_true = re.compile(r'([Yy]es|[Tt]rue)')
bool_false = re.compile(r'([Nn]o|[Ff]alse)')


def _decode_bool(s):
    if bool_true.match(s):
        return True
    if bool_false.match(s):
        return False
    _warn_or_exception(_("Invalid boolean '%s'") % s)


def parse_metadata(metadatapath):
    """Parse metadata file, also checking the source repo for .fdroid.yml.

    If this is a metadata file from fdroiddata, it will first load the
    source repo type and URL from fdroiddata, then read .fdroid.yml if
    it exists, then include the rest of the metadata as specified in
    fdroiddata, so that fdroiddata has precedence over the metadata in
    the source code.

    """
    metadatapath = Path(metadatapath)
    app = App()
    app.metadatapath = metadatapath.as_posix()
    name = metadatapath.stem
    if name != '.fdroid':
        app.id = name

    if metadatapath.suffix == '.yml':
        with metadatapath.open('r', encoding='utf-8') as mf:
            parse_yaml_metadata(mf, app)
    else:
        _warn_or_exception(_('Unknown metadata format: {path} (use: *.yml)')
                           .format(path=metadatapath))

    if metadatapath.name != '.fdroid.yml' and app.Repo:
        build_dir = common.get_build_dir(app)
        metadata_in_repo = build_dir / '.fdroid.yml'
        if metadata_in_repo.is_file():
            try:
                commit_id = common.get_head_commit_id(git.Repo(build_dir))
                logging.debug(_('Including metadata from %s@%s') % (metadata_in_repo, commit_id))
            except git.exc.InvalidGitRepositoryError:
                logging.debug(_('Including metadata from {path}').format(metadata_in_repo))
            app_in_repo = parse_metadata(metadata_in_repo)
            for k, v in app_in_repo.items():
                if k not in app:
                    app[k] = v

    post_metadata_parse(app)

    if not app.id:
        if app.get('Builds'):
            build = app['Builds'][-1]
            if build.subdir:
                root_dir = Path(build.subdir)
            else:
                root_dir = Path('.')
            paths = common.manifest_paths(root_dir, build.gradle)
            _ignored, _ignored, app.id = common.parse_androidmanifests(paths, app)

    return app


def parse_yaml_metadata(mf, app):
    """Parse the .yml file and post-process it.

    Clean metadata .yml files can be used directly, but in order to
    make a better user experience for people editing .yml files, there
    is post processing.  .fdroid.yml is embedded in the app's source
    repo, so it is "user-generated".  That means that it can have
    weird things in it that need to be removed so they don't break the
    overall process.

    """
    try:
        yaml = YAML(typ='safe')
        yamldata = yaml.load(mf)
    except YAMLError as e:
        _warn_or_exception(_("could not parse '{path}'")
                           .format(path=mf.name) + '\n'
                           + common.run_yamllint(mf.name, indent=4),
                           cause=e)

    deprecated_in_yaml = ['Provides']

    if yamldata:
        for field in tuple(yamldata.keys()):
            if field not in yaml_app_fields + deprecated_in_yaml:
                msg = (_("Unrecognised app field '{fieldname}' in '{path}'")
                       .format(fieldname=field, path=mf.name))
                if Path(mf.name).name == '.fdroid.yml':
                    logging.error(msg)
                    del yamldata[field]
                else:
                    _warn_or_exception(msg)

        for deprecated_field in deprecated_in_yaml:
            if deprecated_field in yamldata:
                logging.warning(_("Ignoring '{field}' in '{metapath}' "
                                  "metadata because it is deprecated.")
                                .format(field=deprecated_field,
                                        metapath=mf.name))
                del yamldata[deprecated_field]

        if yamldata.get('Builds', None):
            for build in yamldata.get('Builds', []):
                # put all build flag keywords into a set to avoid
                # excessive looping action
                build_flag_set = set()
                for build_flag in build.keys():
                    build_flag_set.add(build_flag)
                for build_flag in build_flag_set:
                    if build_flag not in build_flags:
                        _warn_or_exception(
                            _("Unrecognised build flag '{build_flag}' "
                              "in '{path}'").format(build_flag=build_flag,
                                                    path=mf.name))
        post_parse_yaml_metadata(yamldata)
        app.update(yamldata)
    return app


def post_parse_yaml_metadata(yamldata):
    """Transform yaml metadata to our internal data format."""
    for build in yamldata.get('Builds', []):
        for flag in build.keys():
            _flagtype = flagtype(flag)

            if _flagtype is TYPE_SCRIPT:
                if isinstance(build[flag], str):
                    build[flag] = [build[flag]]
            elif _flagtype is TYPE_STRING:
                # things like versionNames are strings, but without quotes can be numbers
                if isinstance(build[flag], float) or isinstance(build[flag], int):
                    build[flag] = str(build[flag])
            elif _flagtype is TYPE_INT:
                # versionCode must be int
                if not isinstance(build[flag], int):
                    _warn_or_exception(_('{build_flag} must be an integer, found: {value}')
                                       .format(build_flag=flag, value=build[flag]))


def write_yaml(mf, app):
    """Write metadata in yaml format.

    Parameters
    ----------
    mf
      active file discriptor for writing
    app
      app metadata to written to the yaml file
    """
    # import rumael.yaml and check version
    try:
        import ruamel.yaml
    except ImportError as e:
        raise FDroidException('ruamel.yaml not installed, can not write metadata.') from e
    if not ruamel.yaml.__version__:
        raise FDroidException('ruamel.yaml.__version__ not accessible. Please make sure a ruamel.yaml >= 0.13 is installed..')
    m = re.match(r'(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+)(-.+)?',
                 ruamel.yaml.__version__)
    if not m:
        raise FDroidException('ruamel.yaml version malfored, please install an upstream version of ruamel.yaml')
    if int(m.group('major')) < 0 or int(m.group('minor')) < 13:
        raise FDroidException('currently installed version of ruamel.yaml ({}) is too old, >= 1.13 required.'.format(ruamel.yaml.__version__))
    # suiteable version ruamel.yaml imported successfully

    _yaml_bools_true = ('y', 'Y', 'yes', 'Yes', 'YES',
                        'true', 'True', 'TRUE',
                        'on', 'On', 'ON')
    _yaml_bools_false = ('n', 'N', 'no', 'No', 'NO',
                         'false', 'False', 'FALSE',
                         'off', 'Off', 'OFF')
    _yaml_bools_plus_lists = []
    _yaml_bools_plus_lists.extend(_yaml_bools_true)
    _yaml_bools_plus_lists.extend([[x] for x in _yaml_bools_true])
    _yaml_bools_plus_lists.extend(_yaml_bools_false)
    _yaml_bools_plus_lists.extend([[x] for x in _yaml_bools_false])

    def _field_to_yaml(typ, value):
        if typ is TYPE_STRING:
            if value in _yaml_bools_plus_lists:
                return ruamel.yaml.scalarstring.SingleQuotedScalarString(str(value))
            return str(value)
        elif typ is TYPE_INT:
            return int(value)
        elif typ is TYPE_MULTILINE:
            if '\n' in value:
                return ruamel.yaml.scalarstring.preserve_literal(str(value))
            else:
                return str(value)
        elif typ is TYPE_SCRIPT:
            if type(value) == list:
                if len(value) == 1:
                    return value[0]
                else:
                    return value
        else:
            return value

    def _app_to_yaml(app):
        cm = ruamel.yaml.comments.CommentedMap()
        insert_newline = False
        for field in yaml_app_field_order:
            if field == '\n':
                # next iteration will need to insert a newline
                insert_newline = True
            else:
                if app.get(field) or field == 'Builds':
                    if field == 'Builds':
                        if app.get('Builds'):
                            cm.update({field: _builds_to_yaml(app)})
                    elif field == 'CurrentVersionCode':
                        cm.update({field: _field_to_yaml(TYPE_INT, getattr(app, field))})
                    elif field == 'AllowedAPKSigningKeys':
                        value = getattr(app, field)
                        if value:
                            value = [str(i).lower() for i in value]
                            if len(value) == 1:
                                cm.update({field: _field_to_yaml(TYPE_STRING, value[0])})
                            else:
                                cm.update({field: _field_to_yaml(TYPE_LIST, value)})
                    else:
                        cm.update({field: _field_to_yaml(fieldtype(field), getattr(app, field))})

                    if insert_newline:
                        # we need to prepend a newline in front of this field
                        insert_newline = False
                        # inserting empty lines is not supported so we add a
                        # bogus comment and over-write its value
                        cm.yaml_set_comment_before_after_key(field, 'bogus')
                        cm.ca.items[field][1][-1].value = '\n'
        return cm

    def _builds_to_yaml(app):
        builds = ruamel.yaml.comments.CommentedSeq()
        for build in app.get('Builds', []):
            if not isinstance(build, Build):
                build = Build(build)
            b = ruamel.yaml.comments.CommentedMap()
            for field in build_flags:
                if hasattr(build, field):
                    value = getattr(build, field)
                    if field == 'gradle' and value == ['off']:
                        value = [ruamel.yaml.scalarstring.SingleQuotedScalarString('off')]
                    if field in ('maven'):
                        if value == 'no':
                            continue
                    typ = flagtype(field)
                    # don't check value == True for TYPE_INT as it could be 0
                    if value is not None and (typ == TYPE_INT or value):
                        b.update({field: _field_to_yaml(typ, value)})
            builds.append(b)

        # insert extra empty lines between build entries
        for i in range(1, len(builds)):
            builds.yaml_set_comment_before_after_key(i, 'bogus')
            builds.ca.items[i][1][-1].value = '\n'

        return builds

    yaml_app = _app_to_yaml(app)
    try:
        yaml = ruamel.yaml.YAML()
        yaml.indent(mapping=4, sequence=4, offset=2)
        yaml.dump(yaml_app, stream=mf)
    except AttributeError:  # Debian/stretch's version does not have YAML()
        ruamel.yaml.round_trip_dump(yaml_app, mf, indent=4, block_seq_indent=2)


build_line_sep = re.compile(r'(?<!\\),')
build_cont = re.compile(r'^[ \t]')


def write_metadata(metadatapath, app):
    metadatapath = Path(metadatapath)
    if metadatapath.suffix == '.yml':
        if importlib.util.find_spec('ruamel.yaml'):
            with metadatapath.open('w') as mf:
                return write_yaml(mf, app)
        else:
            raise FDroidException(_('ruamel.yaml not installed, can not write metadata.'))

    _warn_or_exception(_('Unknown metadata format: %s') % metadatapath)


def add_metadata_arguments(parser):
    """Add common command line flags related to metadata processing."""
    parser.add_argument("-W", choices=['error', 'warn', 'ignore'], default='error',
                        help=_("force metadata errors (default) to be warnings, or to be ignored."))



#!/usr/bin/env python3

import ipaddress
import logging
import os
import posixpath
import socket
import subprocess
import sys
from argparse import ArgumentParser
import urllib.parse

from . import _
from . import common
from . import index
from . import update

options = None


def _run_wget(path, urls):
    if options.verbose:
        verbose = '--verbose'
    else:
        verbose = '--no-verbose'

    if not urls:
        return
    logging.debug(_('Running wget in {path}').format(path=path))
    cwd = os.getcwd()
    os.makedirs(path, exist_ok=True)
    os.chdir(path)
    urls_file = '.fdroid-mirror-wget-input-file'
    with open(urls_file, 'w') as fp:
        for url in urls:
            fp.write(url.split('?')[0] + '\n')  # wget puts query string in the filename
    subprocess.call(
        [
            'wget',
            verbose,
            '--continue',
            '--user-agent="fdroid mirror"',
            '--input-file=' + urls_file,
        ]
    )
    os.remove(urls_file)
    os.chdir(cwd)  # leave the working env the way we found it


def main():
    global options

    parser = ArgumentParser()
    common.setup_global_opts(parser)
    parser.add_argument(
        "url",
        nargs='?',
        help=_(
            'Base URL to mirror, can include the index signing key '
            + 'using the query string: ?fingerprint='
        ),
    )
    parser.add_argument(
        "--all",
        action='store_true',
        default=False,
        help=_("Mirror the full repo and archive, all file types."),
    )
    parser.add_argument(
        "--archive",
        action='store_true',
        default=False,
        help=_("Also mirror the full archive section"),
    )
    parser.add_argument(
        "--build-logs",
        action='store_true',
        default=False,
        help=_("Include the build logs in the mirror"),
    )
    parser.add_argument(
        "--pgp-signatures",
        action='store_true',
        default=False,
        help=_("Include the PGP signature .asc files in the mirror"),
    )
    parser.add_argument(
        "--src-tarballs",
        action='store_true',
        default=False,
        help=_("Include the source tarballs in the mirror"),
    )
    parser.add_argument(
        "--output-dir", default=None, help=_("The directory to write the mirror to")
    )
    options = parser.parse_args()

    if options.all:
        options.archive = True
        options.build_logs = True
        options.pgp_signatures = True
        options.src_tarballs = True

    if options.url is None:
        logging.error(_('A URL is required as an argument!') + '\n')
        parser.print_help()
        sys.exit(1)

    scheme, hostname, path, params, query, fragment = urllib.parse.urlparse(options.url)
    fingerprint = urllib.parse.parse_qs(query).get('fingerprint')

    def _append_to_url_path(*args):
        """Append the list of path components to URL, keeping the rest the same."""
        newpath = posixpath.join(path, *args)
        return urllib.parse.urlunparse(
            (scheme, hostname, newpath, params, query, fragment)
        )

    if fingerprint:
        config = common.read_config(options)
        if not ('jarsigner' in config or 'apksigner' in config):
            logging.error(
                _('Java JDK not found! Install in standard location or set java_paths!')
            )
            sys.exit(1)

        def _get_index(section, etag=None):
            url = _append_to_url_path(section)
            data, etag = index.download_repo_index(url, etag=etag)
            return data, etag, _append_to_url_path(section, 'index-v1.jar')

    else:

        def _get_index(section, etag=None):
            import io
            import json
            import zipfile
            from . import net

            url = _append_to_url_path(section, 'index-v1.jar')
            content, etag = net.http_get(url)
            with zipfile.ZipFile(io.BytesIO(content)) as zip:
                jsoncontents = zip.open('index-v1.json').read()
            data = json.loads(jsoncontents.decode('utf-8'))
            return data, etag, None  # no verified index file to return

    ip = None
    try:
        ip = ipaddress.ip_address(hostname)
    except ValueError:
        pass
    if hostname == 'f-droid.org' or (
        ip is not None and hostname in socket.gethostbyname_ex('f-droid.org')[2]
    ):
        print(
            _(
                'ERROR: this command should never be used to mirror f-droid.org!\n'
                'A full mirror of f-droid.org requires more than 200GB.'
            )
        )
        sys.exit(1)

    path = path.rstrip('/')
    if path.endswith('repo') or path.endswith('archive'):
        logging.warning(
            _('Do not include "{path}" in URL!').format(path=path.split('/')[-1])
        )
    elif not path.endswith('fdroid'):
        logging.warning(
            _('{url} does not end with "fdroid", check the URL path!').format(
                url=options.url
            )
        )

    icondirs = ['icons']
    for density in update.screen_densities:
        icondirs.append('icons-' + density)

    if options.output_dir:
        basedir = options.output_dir
    else:
        basedir = os.path.join(os.getcwd(), hostname, path.strip('/'))
        os.makedirs(basedir, exist_ok=True)

    if options.archive:
        sections = ('repo', 'archive')
    else:
        sections = ('repo',)

    for section in sections:
        sectiondir = os.path.join(basedir, section)

        urls = []
        data, etag, index_url = _get_index(section)
        if index_url:
            urls.append(index_url)

        os.makedirs(sectiondir, exist_ok=True)
        os.chdir(sectiondir)
        for icondir in icondirs:
            os.makedirs(os.path.join(sectiondir, icondir), exist_ok=True)

        for packageName, packageList in data['packages'].items():
            for package in packageList:
                to_fetch = []
                keys = ['apkName']
                if options.src_tarballs:
                    keys.append('srcname')
                for k in keys:
                    if k in package:
                        to_fetch.append(package[k])
                    elif k == 'apkName':
                        logging.error(
                            _('{appid} is missing {name}').format(
                                appid=package['packageName'], name=k
                            )
                        )
                for f in to_fetch:
                    if not os.path.exists(f) or (
                        f.endswith('.apk') and os.path.getsize(f) != package['size']
                    ):
                        urls.append(_append_to_url_path(section, f))
                        if options.pgp_signatures:
                            urls.append(_append_to_url_path(section, f + '.asc'))
                        if options.build_logs and f.endswith('.apk'):
                            urls.append(
                                _append_to_url_path(section, f[:-4] + '.log.gz')
                            )

        _run_wget(sectiondir, urls)

        for app in data['apps']:
            localized = app.get('localized')
            if localized:
                for locale, d in localized.items():
                    urls = []
                    components = (section, app['packageName'], locale)
                    for k in update.GRAPHIC_NAMES:
                        f = d.get(k)
                        if f:
                            filepath_tuple = components + (f,)
                            urls.append(_append_to_url_path(*filepath_tuple))
                    _run_wget(os.path.join(basedir, *components), urls)
                    for k in update.SCREENSHOT_DIRS:
                        urls = []
                        filelist = d.get(k)
                        if filelist:
                            components = (section, app['packageName'], locale, k)
                            for f in filelist:
                                filepath_tuple = components + (f,)
                                urls.append(_append_to_url_path(*filepath_tuple))
                            _run_wget(os.path.join(basedir, *components), urls)

        urls = dict()
        for app in data['apps']:
            if 'icon' not in app:
                logging.error(
                    _('no "icon" in {appid}').format(appid=app['packageName'])
                )
                continue
            icon = app['icon']
            for icondir in icondirs:
                url = _append_to_url_path(section, icondir, icon)
                if icondir not in urls:
                    urls[icondir] = []
                urls[icondir].append(url)

        for icondir in icondirs:
            _run_wget(os.path.join(basedir, section, icondir), urls[icondir])


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# update.py - part of the FDroid server tools
# Copyright (C) 2017, Torsten Grote <t at grobox dot de>
# Copyright (C) 2016, Blue Jay Wireless
# Copyright (C) 2014-2016, Hans-Christoph Steiner <hans@eds.org>
# Copyright (C) 2010-2015, Ciaran Gultnieks <ciaran@ciarang.com>
# Copyright (C) 2013-2014, Daniel Martí <mvdan@mvdan.cc>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import hashlib
import json
import logging
import os
import re
import shutil
import tempfile
import urllib.parse
import yaml
import zipfile
import calendar
import qrcode
from binascii import hexlify, unhexlify
from datetime import datetime, timezone
from pathlib import Path
from xml.dom.minidom import Document

from . import _
from . import common
from . import metadata
from . import net
from . import signindex
from fdroidserver.common import FDroidPopen, FDroidPopenBytes, load_stats_fdroid_signing_key_fingerprints
from fdroidserver.exception import FDroidException, VerificationException


def make(apps, apks, repodir, archive):
    """Generate the repo index files.

    This requires properly initialized options and config objects.

    Parameters
    ----------
    apps
      OrderedDict of apps to go into the index, each app should have
      at least one associated apk
    apks
      list of apks to go into the index
    repodir
      the repo directory
    archive
      True if this is the archive repo, False if it's the
      main one.
    """
    from fdroidserver.update import METADATA_VERSION

    if not hasattr(common.options, 'nosign') or not common.options.nosign:
        common.assert_config_keystore(common.config)

    # Historically the index has been sorted by App Name, so we enforce this ordering here
    sortedids = sorted(apps, key=lambda appid: common.get_app_display_name(apps[appid]).upper())
    sortedapps = collections.OrderedDict()
    for appid in sortedids:
        sortedapps[appid] = apps[appid]

    repodict = collections.OrderedDict()
    repodict['timestamp'] = datetime.utcnow().replace(tzinfo=timezone.utc)
    repodict['version'] = METADATA_VERSION

    if common.config['repo_maxage'] != 0:
        repodict['maxage'] = common.config['repo_maxage']

    if archive:
        repodict['name'] = common.config['archive_name']
        repodict['icon'] = common.config.get('archive_icon', common.default_config['repo_icon'])
        repodict['description'] = common.config['archive_description']
        archive_url = common.config.get('archive_url', common.config['repo_url'][:-4] + 'archive')
        repodict['address'] = archive_url
        if 'archive_web_base_url' in common.config:
            repodict["webBaseUrl"] = common.config['archive_web_base_url']
        urlbasepath = os.path.basename(urllib.parse.urlparse(archive_url).path)
    else:
        repodict['name'] = common.config['repo_name']
        repodict['icon'] = common.config.get('repo_icon', common.default_config['repo_icon'])
        repodict['address'] = common.config['repo_url']
        if 'repo_web_base_url' in common.config:
            repodict["webBaseUrl"] = common.config['repo_web_base_url']
        repodict['description'] = common.config['repo_description']
        urlbasepath = os.path.basename(urllib.parse.urlparse(common.config['repo_url']).path)

    mirrorcheckfailed = False
    mirrors = []
    for mirror in common.config.get('mirrors', []):
        base = os.path.basename(urllib.parse.urlparse(mirror).path.rstrip('/'))
        if common.config.get('nonstandardwebroot') is not True and base != 'fdroid':
            logging.error(_("mirror '%s' does not end with 'fdroid'!") % mirror)
            mirrorcheckfailed = True
        # must end with / or urljoin strips a whole path segment
        if mirror.endswith('/'):
            mirrors.append(urllib.parse.urljoin(mirror, urlbasepath))
        else:
            mirrors.append(urllib.parse.urljoin(mirror + '/', urlbasepath))
    for mirror in common.config.get('servergitmirrors', []):
        for url in get_mirror_service_urls(mirror):
            mirrors.append(url + '/' + repodir)
    if mirrorcheckfailed:
        raise FDroidException(_("Malformed repository mirrors."))
    if mirrors:
        repodict['mirrors'] = mirrors

    requestsdict = collections.OrderedDict()
    for command in ('install', 'uninstall'):
        packageNames = []
        key = command + '_list'
        if key in common.config:
            if isinstance(common.config[key], str):
                packageNames = [common.config[key]]
            elif all(isinstance(item, str) for item in common.config[key]):
                packageNames = common.config[key]
            else:
                raise TypeError(_('only accepts strings, lists, and tuples'))
        requestsdict[command] = packageNames

    fdroid_signing_key_fingerprints = load_stats_fdroid_signing_key_fingerprints()

    make_v0(sortedapps, apks, repodir, repodict, requestsdict,
            fdroid_signing_key_fingerprints)
    make_v1(sortedapps, apks, repodir, repodict, requestsdict,
            fdroid_signing_key_fingerprints)
    make_v2(sortedapps, apks, repodir, repodict, requestsdict,
            fdroid_signing_key_fingerprints, archive)
    make_website(sortedapps, repodir, repodict)


def _should_file_be_generated(path, magic_string):
    if os.path.exists(path):
        with open(path) as f:
            # if the magic_string is not in the first line the file should be overwritten
            if magic_string not in f.readline():
                return False
    return True


def make_website(apps, repodir, repodict):
    _ignored, repo_pubkey_fingerprint = extract_pubkey()
    repo_pubkey_fingerprint_stripped = repo_pubkey_fingerprint.replace(" ", "")
    link = repodict["address"]
    link_fingerprinted = ('{link}?fingerprint={fingerprint}'
                          .format(link=link, fingerprint=repo_pubkey_fingerprint_stripped))
    # do not change this string, as it will break updates for files with older versions of this string
    autogenerate_comment = "auto-generated - fdroid index updates will overwrite this file"

    if not os.path.exists(repodir):
        os.makedirs(repodir)

    qrcode.make(link_fingerprinted).save(os.path.join(repodir, "index.png"))

    html_name = 'index.html'
    html_file = os.path.join(repodir, html_name)

    if _should_file_be_generated(html_file, autogenerate_comment):
        with open(html_file, 'w') as f:
            name = repodict["name"]
            description = repodict["description"]
            icon = repodict["icon"]
            f.write("""<!-- {autogenerate_comment} -->
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
  <head>
    <meta content="text/html; charset=utf-8" http-equiv="Content-Type">
    <meta content="width=device-width; initial-scale=1.0; minimum-scale=0.5; maximum-scale=2.0; user-scalable=1;" name="viewport">
    <title>
   {name}
    </title>
    <base href="index.html">
    <link href="index.css" rel="stylesheet" type="text/css">
    <link href="icons/{icon}" rel="icon" type="image/png">
    <link href="icons/{icon}" rel="shortcut icon" type="image/png">
    <meta content="{name}" property="og:site_name">
    <meta content="{name}" property="og:title">
    <meta content property="og:determiner">
    <meta content="{description}" property="og:description">
    <meta content="index,nofollow" name="robots">
  </head>
  <body>
    <h2>
      {name}
    </h2>
    <div id="intro">
      <p style="margin-bottom:.2em;">
        <span style="float:right;width:100px;margin-left:.5em;">
          <a href="index.png" title="QR: test">
            <img alt="QR: test" src="index.png" width="100">
          </a>
        </span>
        {description}
        <br>
        <br>
        Currently it serves
        <kbd>
          {number_of_apps}
        </kbd>
        apps. To add it to your F-Droid client, scan the QR code (click it to enlarge) or use this URL:
      </p>
      <p class="center" style="margin-top:.5em">
        <a href="{link_fingerprinted}">
          <code style="color:#000000;font-weight:bold;">
            {link}
          </code>
        </a>
      </p>
      <p>
        If you would like to manually verify the fingerprint (SHA-256) of the repository signing key, here it is:
        <br>
        <blockcode style="color:#000000;font-weight:bold;">
          {fingerprint}
        </blockcode>
      </p>
    </div>
  </body>
</html>
""".format(autogenerate_comment=autogenerate_comment,
                    description=description,
                    fingerprint=repo_pubkey_fingerprint,
                    icon=icon,
                    link=link,
                    link_fingerprinted=link_fingerprinted,
                    name=name,
                    number_of_apps=str(len(apps))))

    css_file = os.path.join(repodir, "index.css")
    if _should_file_be_generated(css_file, autogenerate_comment):
        with open(css_file, "w") as f:
            # this auto generated comment was not included via .format(), as python seems to have problems with css files in combination with .format()
            f.write("""/* auto-generated - fdroid index updates will overwrite this file */
BODY {
  font-family         : Arial, Helvetica, Sans-Serif;
  color               : #0000ee;
  background-color    : #ffffff;
}
p {
  text-align          : justify;
}
p.center {
  text-align          : center;
}
TD {
  font-family         : Arial, Helvetica, Sans-Serif;
  color               : #0000ee;
}
body,td {
  font-size           : 14px;
}
TH {
  font-family         : Arial, Helvetica, Sans-Serif;
  color               : #0000ee;
  background-color    : #F5EAD4;
}
a:link {
  color               : #bb0000;
}
a:visited {
  color               : #ff0000;
}
.zitat {
  margin-left         : 1cm;
  margin-right        : 1cm;
  font-style          : italic;
}
#intro {
  border-spacing      : 1em;
  border              : 1px solid gray;
  border-radius       : 0.5em;
  box-shadow          : 10px 10px 5px #888;
  margin              : 1.5em;
  font-size           : .9em;
  width               : 600px;
  max-width           : 90%;
  display             : table;
  margin-left         : auto;
  margin-right        : auto;
  font-size           : .8em;
  color               : #555555;
}
#intro > p {
  margin-top          : 0;
}
#intro p:last-child {
  margin-bottom       : 0;
}
.last {
  border-bottom       : 1px solid black;
  padding-bottom      : .5em;
  text-align          : center;
}
table {
  border-collapse     : collapse;
}
h2 {
  text-align          : center;
}
.perms {
  font-family         : monospace;
  font-size           : .8em;
}
.repoapplist {
  display             : table;
  border-collapse     : collapse;
  margin-left         : auto;
  margin-right        : auto;
  width               : 600px;
  max-width           : 90%;
}
.approw, appdetailrow {
  display             : table-row;
}
.appdetailrow {
  display             : flex;
  padding             : .5em;
}
.appiconbig, .appdetailblock, .appdetailcell {
  display             : table-cell
}
.appiconbig {
  vertical-align      : middle;
  text-align          : center;
}
.appdetailinner {
  width               : 100%;
}
.applinkcell {
  text-align          : center;
  float               : right;
  width               : 100%;
  margin-bottom       : .1em;
}
.paddedlink {
  margin              : 1em;
}
.approw {
  border-spacing      : 1em;
  border              : 1px solid gray;
  border-radius       : 0.5em;
  padding             : 0.5em;
  margin              : 1.5em;
}
.appdetailinner .appdetailrow:first-child {
  background-color    : #d5d5d5;
}
.appdetailinner .appdetailrow:first-child .appdetailcell {
  min-width           : 33%;
  flex                : 1 33%;
  text-align          : center;
}
.appdetailinner .appdetailrow:first-child .appdetailcell:first-child {
  text-align          : left;
}
.appdetailinner .appdetailrow:first-child .appdetailcell:last-child {
  float               : none;
  text-align          : right;
}
.minor-details {
  font-size           : .8em;
  color               : #555555;
}
.boldname {
  font-weight         : bold;
}
#appcount {
  text-align          : center;
  margin-bottom       : .5em;
}
kbd {
  padding             : 0.1em 0.6em;
  border              : 1px solid #CCC;
  background-color    : #F7F7F7;
  color               : #333;
  box-shadow          : 0px 1px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 2px #FFF inset;
  border-radius       : 3px;
  display             : inline-block;
  margin              : 0px 0.1em;
  text-shadow         : 0px 1px 0px #FFF;
  white-space         : nowrap;
}
div.filterline, div.repoline {
  display             : table;
  margin-left         : auto;
  margin-right        : auto;
  margin-bottom       : 1em;
  vertical-align      : middle;
  display             : table;
  font-size           : .8em;
}
.filterline form {
  display             : table-row;
}
.filterline .filtercell {
  display             : table-cell;
  vertical-align      : middle;
}
fieldset {
  float               : left;
}
fieldset select, fieldset input, #reposelect select, #reposelect input {
  font-size           : .9em;
}
.pager {
  display             : table;
  margin-left         : auto;
  margin-right        : auto;
  width               : 600px;
  max-width           : 90%;
  padding-top         : .6em;
}
/* should correspond to .repoapplist */
.pagerrow {
  display             : table-row;
}
.pagercell {
  display             : table-cell;
}
.pagercell.left {
  text-align          : left;
  padding-right       : 1em;
}
.pagercell.middle {
  text-align          : center;
  font-size           : .9em;
  color               : #555;
}
.pagercell.right {
  text-align          : right;
  padding-left        : 1em;
}
.anti {
  color               : peru;
}
.antibold {
  color               : crimson;
}
#footer {
  text-align          : center;
  margin-top          : 1em;
  font-size           : 11px;
  color               : #555;
}
#footer img {
  vertical-align      : middle;
}
@media (max-width: 600px) {
  .repoapplist {
    display             : block;
  }
  .appdetailinner, .appdetailrow {
    display             : block;
  }
  .appdetailcell {
    display             : block;
    float               : left;
    line-height         : 1.5em;
  }
}""")


def dict_diff(source, target):
    if not isinstance(target, dict) or not isinstance(source, dict):
        return target

    result = {key: None for key in source if key not in target}

    for key, value in target.items():
        if key not in source:
            result[key] = value
        elif value != source[key]:
            result[key] = dict_diff(source[key], value)

    return result


def file_entry(filename, hash_value=None):
    meta = {}
    meta["name"] = "/" + filename.split("/", 1)[1]
    meta["sha256"] = hash_value or common.sha256sum(filename)
    meta["size"] = os.stat(filename).st_size
    return meta


def load_locale(name, repodir):
    lst = {}
    for yml in Path().glob("config/**/{name}.yml".format(name=name)):
        locale = yml.parts[1]
        if len(yml.parts) == 2:
            locale = "en-US"
        with open(yml, encoding="utf-8") as fp:
            elem = yaml.safe_load(fp)
            for akey, avalue in elem.items():
                if akey not in lst:
                    lst[akey] = {}
                for key, value in avalue.items():
                    if key not in lst[akey]:
                        lst[akey][key] = {}
                    if key == "icon":
                        shutil.copy(os.path.join("config", value), os.path.join(repodir, "icons"))
                        lst[akey][key][locale] = file_entry(os.path.join(repodir, "icons", value))
                    else:
                        lst[akey][key][locale] = value

    return lst


def convert_datetime(obj):
    if isinstance(obj, datetime):
        # Java prefers milliseconds
        # we also need to account for time zone/daylight saving time
        return int(calendar.timegm(obj.timetuple()) * 1000)
    return obj


def package_metadata(app, repodir):
    meta = {}
    for element in (
        "added",
        # "binaries",
        "Categories",
        "Changelog",
        "IssueTracker",
        "lastUpdated",
        "License",
        "SourceCode",
        "Translation",
        "WebSite",
        "featureGraphic",
        "promoGraphic",
        "tvBanner",
        "screenshots",
        "AuthorEmail",
        "AuthorName",
        "AuthorPhone",
        "AuthorWebSite",
        "Bitcoin",
        "FlattrID",
        "Liberapay",
        "Litecoin",
        "OpenCollective",
    ):
        if element in app and app[element]:
            element_new = element[:1].lower() + element[1:]
            meta[element_new] = convert_datetime(app[element])

    for element in (
        "Name",
        "Summary",
        "Description",
        "video",
    ):
        element_new = element[:1].lower() + element[1:]
        if element in app and app[element]:
            meta[element_new] = {"en-US": convert_datetime(app[element])}
        elif "localized" in app:
            localized = {k: v[element_new] for k, v in app["localized"].items() if element_new in v}
            if localized:
                meta[element_new] = localized

    if "name" not in meta and app["AutoName"]:
        meta["name"] = {"en-US": app["AutoName"]}

    # fdroidserver/metadata.py App default
    if meta["license"] == "Unknown":
        del meta["license"]

    if app["Donate"]:
        meta["donate"] = [app["Donate"]]

    # TODO handle different resolutions
    if app.get("icon"):
        meta["icon"] = {"en-US": file_entry(os.path.join(repodir, "icons", app["icon"]))}

    if "iconv2" in app:
        meta["icon"] = app["iconv2"]

    return meta


def convert_version(version, app, repodir):
    ver = {}
    if "added" in version:
        ver["added"] = convert_datetime(version["added"])
    else:
        ver["added"] = 0

    ver["file"] = {
        "name": "/{}".format(version["apkName"]),
        version["hashType"]: version["hash"],
        "size": version["size"]
    }

    ipfsCIDv1 = version.get("ipfsCIDv1")
    if ipfsCIDv1:
        ver["file"]["ipfsCIDv1"] = ipfsCIDv1

    if "srcname" in version:
        ver["src"] = file_entry(os.path.join(repodir, version["srcname"]))

    if "obbMainFile" in version:
        ver["obbMainFile"] = file_entry(
            os.path.join(repodir, version["obbMainFile"]),
            version["obbMainFileSha256"],
        )

    if "obbPatchFile" in version:
        ver["obbPatchFile"] = file_entry(
            os.path.join(repodir, version["obbPatchFile"]),
            version["obbPatchFileSha256"],
        )

    ver["manifest"] = manifest = {}

    for element in (
        "nativecode",
        "versionName",
        "maxSdkVersion",
    ):
        if element in version:
            manifest[element] = version[element]

    if "versionCode" in version:
        manifest["versionCode"] = version["versionCode"]

    if "features" in version and version["features"]:
        manifest["features"] = features = []
        for feature in version["features"]:
            # TODO get version from manifest, default (0) is omitted
            # features.append({"name": feature, "version": 1})
            features.append({"name": feature})

    if "minSdkVersion" in version:
        manifest["usesSdk"] = {}
        manifest["usesSdk"]["minSdkVersion"] = version["minSdkVersion"]
        if "targetSdkVersion" in version:
            manifest["usesSdk"]["targetSdkVersion"] = version["targetSdkVersion"]
        else:
            # https://developer.android.com/guide/topics/manifest/uses-sdk-element.html#target
            manifest["usesSdk"]["targetSdkVersion"] = manifest["usesSdk"]["minSdkVersion"]

    if "signer" in version:
        manifest["signer"] = {"sha256": [version["signer"]]}

    for element in ("uses-permission", "uses-permission-sdk-23"):
        en = element.replace("uses-permission", "usesPermission").replace("-sdk-23", "Sdk23")
        if element in version and version[element]:
            manifest[en] = []
            for perm in version[element]:
                if perm[1]:
                    manifest[en].append({"name": perm[0], "maxSdkVersion": perm[1]})
                else:
                    manifest[en].append({"name": perm[0]})

    if "AntiFeatures" in app and app["AntiFeatures"]:
        ver["antiFeatures"] = {}
        for antif in app["AntiFeatures"]:
            # TODO: get reasons from fdroiddata
            # ver["antiFeatures"][antif] = {"en-US": "reason"}
            ver["antiFeatures"][antif] = {}

    if "AntiFeatures" in version and version["AntiFeatures"]:
        if "antiFeatures" not in ver:
            ver["antiFeatures"] = {}
        for antif in version["AntiFeatures"]:
            # TODO: get reasons from fdroiddata
            # ver["antiFeatures"][antif] = {"en-US": "reason"}
            ver["antiFeatures"][antif] = {}

    if "versionCode" in version:
        if version["versionCode"] > app["CurrentVersionCode"]:
            ver["releaseChannels"] = ["Beta"]

    for build in app.get('Builds', []):
        if build['versionCode'] == version['versionCode'] and "whatsNew" in build:
            ver["whatsNew"] = build["whatsNew"]
            break

    return ver


def v2_repo(repodict, repodir, archive):
    repo = {}

    repo["name"] = {"en-US": repodict["name"]}
    repo["description"] = {"en-US": repodict["description"]}
    repo["icon"] = {"en-US": file_entry("{}/icons/{}".format(repodir, repodict["icon"]))}

    config = load_locale("config", repodir)
    if config:
        repo["name"] = config["archive" if archive else "repo"]["name"]
        repo["description"] = config["archive" if archive else "repo"]["description"]
        repo["icon"] = config["archive" if archive else "repo"]["icon"]

    repo["address"] = repodict["address"]
    if "webBaseUrl" in repodict:
        repo["webBaseUrl"] = repodict["webBaseUrl"]

    if "mirrors" in repodict:
        repo["mirrors"] = [{"url": mirror} for mirror in repodict["mirrors"]]

        # the first entry is traditionally the primary mirror
        if repodict['address'] not in repodict["mirrors"]:
            repo["mirrors"].insert(0, {"url": repodict['address'], "isPrimary": True})

    repo["timestamp"] = repodict["timestamp"]

    antiFeatures = load_locale("antiFeatures", repodir)
    if antiFeatures:
        repo["antiFeatures"] = antiFeatures

    categories = load_locale("categories", repodir)
    if categories:
        repo["categories"] = categories

    channels = load_locale("channels", repodir)
    if channels:
        repo["releaseChannels"] = channels

    return repo


def make_v2(apps, packages, repodir, repodict, requestsdict, fdroid_signing_key_fingerprints, archive):

    def _index_encoder_default(obj):
        if isinstance(obj, set):
            return sorted(list(obj))
        if isinstance(obj, datetime):
            # Java prefers milliseconds
            # we also need to account for time zone/daylight saving time
            return int(calendar.timegm(obj.timetuple()) * 1000)
        if isinstance(obj, dict):
            d = collections.OrderedDict()
            for key in sorted(obj.keys()):
                d[key] = obj[key]
            return d
        raise TypeError(repr(obj) + " is not JSON serializable")

    output = collections.OrderedDict()
    output["repo"] = v2_repo(repodict, repodir, archive)
    if requestsdict and requestsdict["install"] or requestsdict["uninstall"]:
        output["repo"]["requests"] = requestsdict

    # establish sort order of the index
    v1_sort_packages(packages, fdroid_signing_key_fingerprints)

    output_packages = collections.OrderedDict()
    output['packages'] = output_packages
    for package in packages:
        packageName = package['packageName']
        if packageName not in apps:
            logging.info(_('Ignoring package without metadata: ') + package['apkName'])
            continue
        if not package.get('versionName'):
            app = apps[packageName]
            for build in app.get('Builds', []):
                if build['versionCode'] == package['versionCode']:
                    versionName = build.get('versionName')
                    logging.info(_('Overriding blank versionName in {apkfilename} from metadata: {version}')
                                 .format(apkfilename=package['apkName'], version=versionName))
                    package['versionName'] = versionName
                    break
        if packageName in output_packages:
            packagelist = output_packages[packageName]
        else:
            packagelist = {}
            output_packages[packageName] = packagelist
            packagelist["metadata"] = package_metadata(apps[packageName], repodir)
            if "signer" in package:
                packagelist["metadata"]["preferredSigner"] = package["signer"]

            packagelist["versions"] = {}

        packagelist["versions"][package["hash"]] = convert_version(package, apps[packageName], repodir)

    entry = {}
    entry["timestamp"] = repodict["timestamp"]

    entry["version"] = repodict["version"]
    if "maxage" in repodict:
        entry["maxAge"] = repodict["maxage"]

    json_name = 'index-v2.json'
    index_file = os.path.join(repodir, json_name)
    with open(index_file, "w", encoding="utf-8") as fp:
        if common.options.pretty:
            json.dump(output, fp, default=_index_encoder_default, indent=2, ensure_ascii=False)
        else:
            json.dump(output, fp, default=_index_encoder_default, ensure_ascii=False)

    json_name = "tmp/{}_{}.json".format(repodir, convert_datetime(repodict["timestamp"]))
    with open(json_name, "w", encoding="utf-8") as fp:
        if common.options.pretty:
            json.dump(output, fp, default=_index_encoder_default, indent=2, ensure_ascii=False)
        else:
            json.dump(output, fp, default=_index_encoder_default, ensure_ascii=False)

    entry["index"] = file_entry(index_file)
    entry["index"]["numPackages"] = len(output.get("packages", []))

    indexes = sorted(Path().glob("tmp/{}*.json".format(repodir)), key=lambda x: x.name)
    indexes.pop()  # remove current index
    # remove older indexes
    while len(indexes) > 10:
        indexes.pop(0).unlink()

    indexes = [json.loads(Path(fn).read_text(encoding="utf-8")) for fn in indexes]

    for diff in Path().glob("{}/diff/*.json".format(repodir)):
        diff.unlink()

    entry["diffs"] = {}
    for old in indexes:
        diff_name = str(old["repo"]["timestamp"]) + ".json"
        diff_file = os.path.join(repodir, "diff", diff_name)
        diff = dict_diff(old, output)
        if not os.path.exists(os.path.join(repodir, "diff")):
            os.makedirs(os.path.join(repodir, "diff"))
        with open(diff_file, "w", encoding="utf-8") as fp:
            if common.options.pretty:
                json.dump(diff, fp, default=_index_encoder_default, indent=2, ensure_ascii=False)
            else:
                json.dump(diff, fp, default=_index_encoder_default, ensure_ascii=False)

        entry["diffs"][old["repo"]["timestamp"]] = file_entry(diff_file)
        entry["diffs"][old["repo"]["timestamp"]]["numPackages"] = len(diff.get("packages", []))

    json_name = "entry.json"
    index_file = os.path.join(repodir, json_name)
    with open(index_file, "w", encoding="utf-8") as fp:
        if common.options.pretty:
            json.dump(entry, fp, default=_index_encoder_default, indent=2, ensure_ascii=False)
        else:
            json.dump(entry, fp, default=_index_encoder_default, ensure_ascii=False)

    if common.options.nosign:
        _copy_to_local_copy_dir(repodir, index_file)
        logging.debug(_('index-v2 must have a signature, use `fdroid signindex` to create it!'))
    else:
        signindex.config = common.config
        signindex.sign_index(repodir, json_name)


def make_v1(apps, packages, repodir, repodict, requestsdict, fdroid_signing_key_fingerprints):

    def _index_encoder_default(obj):
        if isinstance(obj, set):
            return sorted(list(obj))
        if isinstance(obj, datetime):
            # Java prefers milliseconds
            # we also need to account for time zone/daylight saving time
            return int(calendar.timegm(obj.timetuple()) * 1000)
        if isinstance(obj, dict):
            d = collections.OrderedDict()
            for key in sorted(obj.keys()):
                d[key] = obj[key]
            return d
        raise TypeError(repr(obj) + " is not JSON serializable")

    output = collections.OrderedDict()
    output['repo'] = repodict
    output['requests'] = requestsdict

    # establish sort order of the index
    v1_sort_packages(packages, fdroid_signing_key_fingerprints)

    appslist = []
    output['apps'] = appslist
    for packageName, appdict in apps.items():
        d = collections.OrderedDict()
        appslist.append(d)
        for k, v in sorted(appdict.items()):
            if not v:
                continue
            if k in ('Builds', 'comments', 'metadatapath',
                     'ArchivePolicy', 'AutoName', 'AutoUpdateMode', 'MaintainerNotes',
                     'Provides', 'Repo', 'RepoType', 'RequiresRoot',
                     'UpdateCheckData', 'UpdateCheckIgnore', 'UpdateCheckMode',
                     'UpdateCheckName', 'NoSourceSince', 'VercodeOperation',
                     'summary', 'description', 'promoGraphic', 'screenshots', 'whatsNew',
                     'featureGraphic', 'iconv2', 'tvBanner',
                     ):
                continue

            # name things after the App class fields in fdroidclient
            if k == 'id':
                k = 'packageName'
            elif k == 'CurrentVersionCode':  # TODO make SuggestedVersionCode the canonical name
                k = 'suggestedVersionCode'
                v = str(v)
            elif k == 'CurrentVersion':  # TODO make SuggestedVersionName the canonical name
                k = 'suggestedVersionName'
            else:
                k = k[:1].lower() + k[1:]
            d[k] = v

    # establish sort order in lists, sets, and localized dicts
    for app in output['apps']:
        localized = app.get('localized')
        if localized:
            lordered = collections.OrderedDict()
            for lkey, lvalue in sorted(localized.items()):
                lordered[lkey] = collections.OrderedDict()
                for ikey, iname in sorted(lvalue.items()):
                    lordered[lkey][ikey] = iname
            app['localized'] = lordered
        antiFeatures = app.get('AntiFeatures')
        if antiFeatures:
            app['AntiFeatures'] = sorted(set(antiFeatures))

    output_packages = collections.OrderedDict()
    output['packages'] = output_packages
    for package in packages:
        packageName = package['packageName']
        if packageName not in apps:
            logging.info(_('Ignoring package without metadata: ') + package['apkName'])
            continue
        if not package.get('versionName'):
            app = apps[packageName]
            for build in app.get('Builds', []):
                if build['versionCode'] == package['versionCode']:
                    versionName = build.get('versionName')
                    logging.info(_('Overriding blank versionName in {apkfilename} from metadata: {version}')
                                 .format(apkfilename=package['apkName'], version=versionName))
                    package['versionName'] = versionName
                    break
        if packageName in output_packages:
            packagelist = output_packages[packageName]
        else:
            packagelist = []
            output_packages[packageName] = packagelist
        d = collections.OrderedDict()
        packagelist.append(d)
        for k, v in sorted(package.items()):
            if not v:
                continue
            if k in ('icon', 'icons', 'icons_src', 'ipfsCIDv1', 'name'):
                continue
            d[k] = v

    json_name = 'index-v1.json'
    index_file = os.path.join(repodir, json_name)
    with open(index_file, 'w') as fp:
        if common.options.pretty:
            json.dump(output, fp, default=_index_encoder_default, indent=2)
        else:
            json.dump(output, fp, default=_index_encoder_default)

    if common.options.nosign:
        _copy_to_local_copy_dir(repodir, index_file)
        logging.debug(_('index-v1 must have a signature, use `fdroid signindex` to create it!'))
    else:
        signindex.config = common.config
        signindex.sign_index(repodir, json_name)


def _copy_to_local_copy_dir(repodir, f):
    local_copy_dir = common.config.get('local_copy_dir', '')
    if os.path.exists(local_copy_dir):
        destdir = os.path.join(local_copy_dir, repodir)
        if not os.path.exists(destdir):
            os.mkdir(destdir)
        shutil.copy2(f, destdir, follow_symlinks=False)
    elif local_copy_dir:
        raise FDroidException(_('"local_copy_dir" {path} does not exist!')
                              .format(path=local_copy_dir))


def v1_sort_packages(packages, fdroid_signing_key_fingerprints):
    """Sort the supplied list to ensure a deterministic sort order for package entries in the index file.

    This sort-order also expresses
    installation preference to the clients.
    (First in this list = first to install)

    Parameters
    ----------
    packages
      list of packages which need to be sorted before but into index file.
    """
    GROUP_DEV_SIGNED = 1
    GROUP_FDROID_SIGNED = 2
    GROUP_OTHER_SIGNED = 3

    def v1_sort_keys(package):
        packageName = package.get('packageName', None)

        signer = package.get('signer', None)

        dev_signer = common.metadata_find_developer_signature(packageName)
        group = GROUP_OTHER_SIGNED
        if dev_signer and dev_signer == signer:
            group = GROUP_DEV_SIGNED
        else:
            fdroid_signer = fdroid_signing_key_fingerprints.get(packageName, {}).get('signer')
            if fdroid_signer and fdroid_signer == signer:
                group = GROUP_FDROID_SIGNED

        versionCode = None
        if package.get('versionCode', None):
            versionCode = -package['versionCode']

        return packageName, group, signer, versionCode

    packages.sort(key=v1_sort_keys)


def make_v0(apps, apks, repodir, repodict, requestsdict, fdroid_signing_key_fingerprints):
    """Aka index.jar aka index.xml."""
    doc = Document()

    def addElement(name, value, doc, parent):
        el = doc.createElement(name)
        el.appendChild(doc.createTextNode(value))
        parent.appendChild(el)

    def addElementNonEmpty(name, value, doc, parent):
        if not value:
            return
        addElement(name, value, doc, parent)

    def addElementIfInApk(name, apk, key, doc, parent):
        if key not in apk:
            return
        value = str(apk[key])
        addElement(name, value, doc, parent)

    def addElementCheckLocalized(name, app, key, doc, parent, default=''):
        """Fill in field from metadata or localized block.

        For name/summary/description, they can come only from the app source,
        or from a dir in fdroiddata.  They can be entirely missing from the
        metadata file if there is localized versions.  This will fetch those
        from the localized version if its not available in the metadata file.

        Attributes should be alpha-sorted, so they must be added in
        alpha- sort order.

        """
        el = doc.createElement(name)
        value = app.get(key)
        lkey = key[:1].lower() + key[1:]
        localized = app.get('localized')
        if not value and localized:
            for lang in ['en-US'] + [x for x in localized.keys()]:
                if not lang.startswith('en'):
                    continue
                if lang in localized:
                    value = localized[lang].get(lkey)
                    if value:
                        break
        if not value and localized and len(localized) > 1:
            lang = list(localized.keys())[0]
            value = localized[lang].get(lkey)
        if not value:
            value = default
        if not value and name == 'name' and app.get('AutoName'):
            value = app['AutoName']
        el.appendChild(doc.createTextNode(value))
        parent.appendChild(el)

    root = doc.createElement("fdroid")
    doc.appendChild(root)

    repoel = doc.createElement("repo")
    repoel.setAttribute("icon", repodict['icon'])
    if 'maxage' in repodict:
        repoel.setAttribute("maxage", str(repodict['maxage']))
    repoel.setAttribute("name", repodict['name'])
    pubkey, repo_pubkey_fingerprint = extract_pubkey()
    repoel.setAttribute("pubkey", pubkey.decode('utf-8'))
    repoel.setAttribute("timestamp", '%d' % repodict['timestamp'].timestamp())
    repoel.setAttribute("url", repodict['address'])
    repoel.setAttribute("version", str(repodict['version']))

    addElement('description', repodict['description'], doc, repoel)
    for mirror in repodict.get('mirrors', []):
        addElement('mirror', mirror, doc, repoel)

    root.appendChild(repoel)

    for command in ('install', 'uninstall'):
        for packageName in requestsdict[command]:
            element = doc.createElement(command)
            root.appendChild(element)
            element.setAttribute('packageName', packageName)

    for appid, appdict in apps.items():
        app = metadata.App(appdict)

        if app.get('Disabled') is not None:
            continue

        # Get a list of the apks for this app...
        apklist = []
        name_from_apk = None
        apksbyversion = collections.defaultdict(lambda: [])
        for apk in apks:
            if apk.get('versionCode') and apk.get('packageName') == appid:
                apksbyversion[apk['versionCode']].append(apk)
                if name_from_apk is None:
                    name_from_apk = apk.get('name')
        for versionCode, apksforver in apksbyversion.items():
            fdroid_signer = fdroid_signing_key_fingerprints.get(appid, {}).get('signer')
            fdroid_signed_apk = None
            name_match_apk = None
            for x in apksforver:
                if fdroid_signer and x.get('signer', None) == fdroid_signer:
                    fdroid_signed_apk = x
                if common.apk_release_filename.match(x.get('apkName', '')):
                    name_match_apk = x
            # choose which of the available versions is most
            # suiteable for index v0
            if fdroid_signed_apk:
                apklist.append(fdroid_signed_apk)
            elif name_match_apk:
                apklist.append(name_match_apk)
            else:
                apklist.append(apksforver[0])

        if len(apklist) == 0:
            continue

        apel = doc.createElement("application")
        apel.setAttribute("id", app.id)
        root.appendChild(apel)

        addElement('id', app.id, doc, apel)
        if app.added:
            addElement('added', app.added.strftime('%Y-%m-%d'), doc, apel)
        if app.lastUpdated:
            addElement('lastupdated', app.lastUpdated.strftime('%Y-%m-%d'), doc, apel)

        addElementCheckLocalized('name', app, 'Name', doc, apel, name_from_apk)
        addElementCheckLocalized('summary', app, 'Summary', doc, apel)

        if app.icon:
            addElement('icon', app.icon, doc, apel)

        addElementCheckLocalized('desc', app, 'Description', doc, apel,
                                 'No description available')

        addElement('license', app.License, doc, apel)
        if app.Categories:
            addElement('categories', ','.join(app.Categories), doc, apel)
            # We put the first (primary) category in LAST, which will have
            # the desired effect of making clients that only understand one
            # category see that one.
            addElement('category', app.Categories[0], doc, apel)
        addElement('web', app.WebSite, doc, apel)
        addElement('source', app.SourceCode, doc, apel)
        addElement('tracker', app.IssueTracker, doc, apel)
        addElementNonEmpty('changelog', app.Changelog, doc, apel)
        addElementNonEmpty('author', app.AuthorName, doc, apel)
        addElementNonEmpty('email', app.AuthorEmail, doc, apel)
        addElementNonEmpty('donate', app.Donate, doc, apel)
        addElementNonEmpty('bitcoin', app.Bitcoin, doc, apel)
        addElementNonEmpty('litecoin', app.Litecoin, doc, apel)
        addElementNonEmpty('flattr', app.FlattrID, doc, apel)
        addElementNonEmpty('openCollective', app.OpenCollective, doc, apel)

        # These elements actually refer to the current version (i.e. which
        # one is recommended. They are historically mis-named, and need
        # changing, but stay like this for now to support existing clients.
        addElement('marketversion', app.CurrentVersion, doc, apel)
        addElement('marketvercode', str(app.CurrentVersionCode), doc, apel)

        if app.Provides:
            pv = app.Provides.split(',')
            addElementNonEmpty('provides', ','.join(pv), doc, apel)
        if app.RequiresRoot:
            addElement('requirements', 'root', doc, apel)

        # Sort the APK list into version order, just so the web site
        # doesn't have to do any work by default...
        apklist = sorted(apklist, key=lambda apk: apk['versionCode'], reverse=True)

        if 'antiFeatures' in apklist[0]:
            app.AntiFeatures.extend(apklist[0]['antiFeatures'])
        if app.AntiFeatures:
            afout = sorted(set(app.AntiFeatures))
            addElementNonEmpty('antifeatures', ','.join(afout), doc, apel)

        # Check for duplicates - they will make the client unhappy...
        for i in range(len(apklist) - 1):
            first = apklist[i]
            second = apklist[i + 1]
            if first['versionCode'] == second['versionCode'] \
               and first['sig'] == second['sig']:
                if first['hash'] == second['hash']:
                    raise FDroidException('"{0}/{1}" and "{0}/{2}" are exact duplicates!'.format(
                        repodir, first['apkName'], second['apkName']))
                else:
                    raise FDroidException('duplicates: "{0}/{1}" - "{0}/{2}"'.format(
                        repodir, first['apkName'], second['apkName']))

        current_version_code = 0
        current_version_file = None
        for apk in apklist:
            file_extension = common.get_file_extension(apk['apkName'])
            # find the APK for the "Current Version"
            if current_version_code < app.CurrentVersionCode:
                current_version_file = apk['apkName']
            if current_version_code < apk['versionCode']:
                current_version_code = apk['versionCode']

            apkel = doc.createElement("package")
            apel.appendChild(apkel)

            versionName = apk.get('versionName')
            if not versionName:
                for build in app.get('Builds', []):
                    if (
                        build['versionCode'] == apk['versionCode']
                        and 'versionName' in build
                    ):
                        versionName = build['versionName']
                        break
            if versionName:
                addElement('version', versionName, doc, apkel)

            addElement('versioncode', str(apk['versionCode']), doc, apkel)
            addElement('apkname', apk['apkName'], doc, apkel)
            addElementIfInApk('srcname', apk, 'srcname', doc, apkel)

            hashel = doc.createElement("hash")
            hashel.setAttribute('type', 'sha256')
            hashel.appendChild(doc.createTextNode(apk['hash']))
            apkel.appendChild(hashel)

            addElement('size', str(apk['size']), doc, apkel)
            addElementIfInApk('sdkver', apk,
                              'minSdkVersion', doc, apkel)
            addElementIfInApk('targetSdkVersion', apk,
                              'targetSdkVersion', doc, apkel)
            addElementIfInApk('maxsdkver', apk,
                              'maxSdkVersion', doc, apkel)
            addElementIfInApk('obbMainFile', apk,
                              'obbMainFile', doc, apkel)
            addElementIfInApk('obbMainFileSha256', apk,
                              'obbMainFileSha256', doc, apkel)
            addElementIfInApk('obbPatchFile', apk,
                              'obbPatchFile', doc, apkel)
            addElementIfInApk('obbPatchFileSha256', apk,
                              'obbPatchFileSha256', doc, apkel)
            if 'added' in apk:
                addElement('added', apk['added'].strftime('%Y-%m-%d'), doc, apkel)

            if file_extension == 'apk':  # sig is required for APKs, but only APKs
                addElement('sig', apk['sig'], doc, apkel)

                old_permissions = set()
                sorted_permissions = sorted(apk['uses-permission'])
                for perm in sorted_permissions:
                    perm_name = perm[0]
                    if perm_name.startswith("android.permission."):
                        perm_name = perm_name[19:]
                    old_permissions.add(perm_name)
                addElementNonEmpty('permissions', ','.join(sorted(old_permissions)), doc, apkel)

                for permission in sorted_permissions:
                    permel = doc.createElement('uses-permission')
                    if permission[1] is not None:
                        permel.setAttribute('maxSdkVersion', '%d' % permission[1])
                        apkel.appendChild(permel)
                    permel.setAttribute('name', permission[0])
                for permission_sdk_23 in sorted(apk['uses-permission-sdk-23']):
                    permel = doc.createElement('uses-permission-sdk-23')
                    if permission_sdk_23[1] is not None:
                        permel.setAttribute('maxSdkVersion', '%d' % permission_sdk_23[1])
                        apkel.appendChild(permel)
                    permel.setAttribute('name', permission_sdk_23[0])
                if 'nativecode' in apk:
                    addElement('nativecode', ','.join(sorted(apk['nativecode'])), doc, apkel)
                addElementNonEmpty('features', ','.join(sorted(apk['features'])), doc, apkel)

        if current_version_file is not None \
                and common.config['make_current_version_link'] \
                and repodir == 'repo':  # only create these
            namefield = common.config['current_version_name_source']
            name = app.get(namefield)
            if not name and namefield == 'Name':
                name = app.get('localized', {}).get('en-US', {}).get('name')
            if not name:
                name = app.id
            sanitized_name = re.sub(b'''[ '"&%?+=/]''', b'', str(name).encode('utf-8'))
            apklinkname = sanitized_name + os.path.splitext(current_version_file)[1].encode('utf-8')
            current_version_path = os.path.join(repodir, current_version_file).encode('utf-8', 'surrogateescape')
            if os.path.islink(apklinkname):
                os.remove(apklinkname)
            os.symlink(current_version_path, apklinkname)
            # also symlink gpg signature, if it exists
            for extension in (b'.asc', b'.sig'):
                sigfile_path = current_version_path + extension
                if os.path.exists(sigfile_path):
                    siglinkname = apklinkname + extension
                    if os.path.islink(siglinkname):
                        os.remove(siglinkname)
                    os.symlink(sigfile_path, siglinkname)

    if common.options.pretty:
        output = doc.toprettyxml(encoding='utf-8')
    else:
        output = doc.toxml(encoding='utf-8')

    with open(os.path.join(repodir, 'index.xml'), 'wb') as f:
        f.write(output)

    if 'repo_keyalias' in common.config \
       or (common.options.nosign and 'repo_pubkey' in common.config):

        if common.options.nosign:
            logging.info(_("Creating unsigned index in preparation for signing"))
        else:
            logging.info(_("Creating signed index with this key (SHA256):"))
            logging.info("%s" % repo_pubkey_fingerprint)

        # Create a jar of the index...
        jar_output = 'index_unsigned.jar' if common.options.nosign else 'index.jar'
        p = FDroidPopen(['jar', 'cf', jar_output, 'index.xml'], cwd=repodir)
        if p.returncode != 0:
            raise FDroidException("Failed to create {0}".format(jar_output))

        # Sign the index...
        signed = os.path.join(repodir, 'index.jar')
        if common.options.nosign:
            _copy_to_local_copy_dir(repodir, os.path.join(repodir, jar_output))
            # Remove old signed index if not signing
            if os.path.exists(signed):
                os.remove(signed)
        else:
            signindex.config = common.config
            signindex.sign_jar(signed, use_old_algs=True)

    # Copy the repo icon into the repo directory...
    icon_dir = os.path.join(repodir, 'icons')
    repo_icon = common.config.get('repo_icon', common.default_config['repo_icon'])
    iconfilename = os.path.join(icon_dir, os.path.basename(repo_icon))
    if os.path.exists(repo_icon):
        shutil.copyfile(common.config['repo_icon'], iconfilename)
    else:
        logging.warning(_('repo_icon "repo/icons/%s" does not exist, generating placeholder.')
                        % repo_icon)
        os.makedirs(os.path.dirname(iconfilename), exist_ok=True)
        try:
            qrcode.make(common.config['repo_url']).save(iconfilename)
        except Exception:
            exampleicon = os.path.join(common.get_examples_dir(),
                                       common.default_config['repo_icon'])
            shutil.copy(exampleicon, iconfilename)


def extract_pubkey():
    """Extract and return the repository's public key from the keystore.

    Returns
    -------
    public key in hex
    repository fingerprint
    """
    if 'repo_pubkey' in common.config:
        pubkey = unhexlify(common.config['repo_pubkey'])
    elif 'keystorepass' in common.config:
        env_vars = {'LC_ALL': 'C.UTF-8',
                    'FDROID_KEY_STORE_PASS': common.config['keystorepass']}
        p = FDroidPopenBytes([common.config['keytool'], '-exportcert',
                              '-alias', common.config['repo_keyalias'],
                              '-keystore', common.config['keystore'],
                              '-storepass:env', 'FDROID_KEY_STORE_PASS']
                             + list(common.config['smartcardoptions']),
                             envs=env_vars, output=False, stderr_to_stdout=False)
        if p.returncode != 0 or len(p.output) < 20:
            msg = "Failed to get repo pubkey!"
            if common.config['keystore'] == 'NONE':
                msg += ' Is your crypto smartcard plugged in?'
            raise FDroidException(msg)
        pubkey = p.output
    else:
        raise FDroidException(_('Neither "repo_pubkey" nor "keystorepass" set in config.yml'))

    repo_pubkey_fingerprint = common.get_cert_fingerprint(pubkey)
    return hexlify(pubkey), repo_pubkey_fingerprint


def get_mirror_service_urls(url):
    """Get direct URLs from git service for use by fdroidclient.

    Via 'servergitmirrors', fdroidserver can create and push a mirror
    to certain well known git services like GitLab or GitHub.  This
    will always use the 'master' branch since that is the default
    branch in git. The files are then accessible via alternate URLs,
    where they are served in their raw format via a CDN rather than
    from git.

    Both of the GitLab URLs will work with F-Droid, but only the
    GitLab Pages will work in the browser This is because the "raw"
    URLs are not served with the correct mime types, so any index.html
    which is put in the repo will not be rendered. Putting an
    index.html file in the repo root is a common way for to make
    information about the repo available to end user.

    """
    if url.startswith('git@'):
        url = re.sub(r'^git@([^:]+):(.+)', r'https://\1/\2', url)

    segments = url.split("/")

    if segments[4].endswith('.git'):
        segments[4] = segments[4][:-4]

    hostname = segments[2]
    user = segments[3]
    repo = segments[4]
    branch = "master"
    folder = "fdroid"

    urls = []
    if hostname == "github.com":
        # Github-like RAW segments "https://raw.githubusercontent.com/user/repo/branch/folder"
        segments[2] = "raw.githubusercontent.com"
        segments.extend([branch, folder])
        urls.append('/'.join(segments))
    elif hostname == "gitlab.com":
        git_mirror_path = os.path.join('git-mirror', folder)
        if common.get_dir_size(git_mirror_path) <= common.GITLAB_COM_PAGES_MAX_SIZE:
            # Gitlab-like Pages segments "https://user.gitlab.io/repo/folder"
            gitlab_pages = ["https:", "", user + ".gitlab.io", repo, folder]
            urls.append('/'.join(gitlab_pages))
        else:
            logging.warning(
                _(
                    'Skipping GitLab Pages mirror because the repo is too large (>%.2fGB)!'
                )
                % (common.GITLAB_COM_PAGES_MAX_SIZE / 1000000000)
            )
        # GitLab Raw "https://gitlab.com/user/repo/-/raw/branch/folder"
        gitlab_raw = segments + ['-', 'raw', branch, folder]
        urls.append('/'.join(gitlab_raw))
        # GitLab Artifacts "https://user.gitlab.io/-/repo/-/jobs/job_id/artifacts/public/folder"
        job_id = os.getenv('CI_JOB_ID')
        try:
            int(job_id)
            gitlab_artifacts = [
                "https:",
                "",
                user + ".gitlab.io",
                '-',
                repo,
                '-',
                'jobs',
                job_id,
                'artifacts',
                'public',
                folder,
            ]
            urls.append('/'.join(gitlab_artifacts))
        except (TypeError, ValueError):
            pass  # no Job ID to use, ignore

    return urls


def download_repo_index(url_str, etag=None, verify_fingerprint=True, timeout=600):
    """Download and verifies index v1 file, then returns its data.

    Use the versioned functions to be sure you are getting the
    expected data format.

    """
    return download_repo_index_v1(url_str, etag, verify_fingerprint, timeout)


def download_repo_index_v1(url_str, etag=None, verify_fingerprint=True, timeout=600):
    """Download and verifies index v1 file, then returns its data.

    Downloads the repository index from the given :param url_str and
    verifies the repository's fingerprint if :param verify_fingerprint
    is not False.

    Raises
    ------
    VerificationException() if the repository could not be verified

    Returns
    -------
    A tuple consisting of:
        - The index in JSON v1 format or None if the index did not change
        - The new eTag as returned by the HTTP request

    """
    url = urllib.parse.urlsplit(url_str)

    fingerprint = None
    if verify_fingerprint:
        query = urllib.parse.parse_qs(url.query)
        if 'fingerprint' not in query:
            raise VerificationException(_("No fingerprint in URL."))
        fingerprint = query['fingerprint'][0]

    if url.path.endswith('/index-v1.jar'):
        path = url.path[:-13].rstrip('/')
    else:
        path = url.path.rstrip('/')

    url = urllib.parse.SplitResult(url.scheme, url.netloc, path + '/index-v1.jar', '', '')
    download, new_etag = net.http_get(url.geturl(), etag, timeout)

    if download is None:
        return None, new_etag

    with tempfile.NamedTemporaryFile() as fp:
        fp.write(download)
        fp.flush()
        index, public_key, public_key_fingerprint = get_index_from_jar(
            fp.name, fingerprint, allow_deprecated=True
        )
        index["repo"]["pubkey"] = hexlify(public_key).decode()
        index["repo"]["fingerprint"] = public_key_fingerprint
        index["apps"] = [metadata.App(app) for app in index["apps"]]
        return index, new_etag


def download_repo_index_v2(url_str, etag=None, verify_fingerprint=True, timeout=600):
    """Download and verifies index v2 file, then returns its data.

    Downloads the repository index from the given :param url_str and
    verifies the repository's fingerprint if :param verify_fingerprint
    is not False.  In order to verify the data, the fingerprint must
    be provided as part of the URL.

    Raises
    ------
    VerificationException() if the repository could not be verified

    Returns
    -------
    A tuple consisting of:
        - The index in JSON v2 format or None if the index did not change
        - The new eTag as returned by the HTTP request

    """
    url = urllib.parse.urlsplit(url_str)

    fingerprint = None
    if verify_fingerprint:
        query = urllib.parse.parse_qs(url.query)
        if 'fingerprint' not in query:
            raise VerificationException(_("No fingerprint in URL."))
        fingerprint = query['fingerprint'][0]

    if url.path.endswith('/entry.jar') or url.path.endswith('/index-v2.json'):
        path = url.path.rsplit('/', 1)[0]
    else:
        path = url.path.rstrip('/')

    url = urllib.parse.SplitResult(url.scheme, url.netloc, path + '/entry.jar', '', '')
    download, new_etag = net.http_get(url.geturl(), etag, timeout)

    if download is None:
        return None, new_etag

    # jarsigner is used to verify the JAR, it requires a file for input
    with tempfile.TemporaryDirectory() as dirname:
        with (Path(dirname) / 'entry.jar').open('wb') as fp:
            fp.write(download)
            fp.flush()
            entry, public_key, fingerprint = get_index_from_jar(fp.name, fingerprint)

    name = entry['index']['name']
    sha256 = entry['index']['sha256']
    url = urllib.parse.SplitResult(url.scheme, url.netloc, path + name, '', '')
    index, _ignored = net.http_get(url.geturl(), None, timeout)
    if sha256 != hashlib.sha256(index).hexdigest():
        raise VerificationException(
            _("SHA-256 of {url} does not match entry!").format(url=url)
        )
    return json.loads(index), new_etag


def get_index_from_jar(jarfile, fingerprint=None, allow_deprecated=False):
    """Return the data, public key and fingerprint from an index JAR with one JSON file.

    The F-Droid index files always contain a single data file and a
    JAR Signature.  Since index-v1, the data file is always JSON.
    That single data file is named the same as the JAR file.

    Parameters
    ----------
    fingerprint is the SHA-256 fingerprint of signing key. Only
      hex digits count, all other chars will can be discarded.

    Raises
    ------
    VerificationException() if the repository could not be verified

    """
    logging.debug(_('Verifying index signature:'))

    if allow_deprecated:
        common.verify_deprecated_jar_signature(jarfile)
    else:
        common.verify_jar_signature(jarfile)

    with zipfile.ZipFile(jarfile) as jar:
        public_key, public_key_fingerprint = get_public_key_from_jar(jar)
        if fingerprint is not None:
            fingerprint = re.sub(r'[^0-9A-F]', r'', fingerprint.upper())
            if fingerprint != public_key_fingerprint:
                raise VerificationException(
                    _("The repository's fingerprint does not match.")
                )
        for f in jar.namelist():
            if not f.startswith('META-INF/'):
                jsonfile = f
                break
        data = json.loads(jar.read(jsonfile))
        return data, public_key, public_key_fingerprint


def get_public_key_from_jar(jar):
    """Get the public key and its fingerprint from a JAR file.

    Raises
    ------
    VerificationException() if the JAR was not signed exactly once

    Parameters
    ----------
    jar
      a zipfile.ZipFile object

    Returns
    -------
    the public key from the jar and its fingerprint
    """
    # extract certificate from jar
    certs = [n for n in jar.namelist() if common.SIGNATURE_BLOCK_FILE_REGEX.match(n)]
    if len(certs) < 1:
        raise VerificationException(_("Found no signing certificates for repository."))
    if len(certs) > 1:
        raise VerificationException(_("Found multiple signing certificates for repository."))

    # extract public key from certificate
    public_key = common.get_certificate(jar.read(certs[0]))
    public_key_fingerprint = common.get_cert_fingerprint(public_key).replace(' ', '')

    return public_key, public_key_fingerprint



#!/usr/bin/env python3
#
# net.py - part of the FDroid server tools
# Copyright (C) 2015 Hans-Christoph Steiner <hans@eds.org>
# Copyright (C) 2022 FC Stegerman <flx@obfusk.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import requests
import time
import urllib
from requests.adapters import HTTPAdapter, Retry
from requests.exceptions import ChunkedEncodingError

HEADERS = {'User-Agent': 'F-Droid'}


def download_file(url, local_filename=None, dldir='tmp', retries=3, backoff_factor=0.1):
    filename = urllib.parse.urlparse(url).path.split('/')[-1]
    if local_filename is None:
        local_filename = os.path.join(dldir, filename)
    # Retry applies to failed DNS lookups, socket connections and connection
    # timeouts, never to requests where data has made it to the server; so we
    # handle ChunkedEncodingError during transfer ourselves.
    for i in range(retries + 1):
        if retries:
            max_retries = Retry(total=retries - i, backoff_factor=backoff_factor)
            adapter = HTTPAdapter(max_retries=max_retries)
            session = requests.Session()
            session.mount('http://', adapter)
            session.mount('https://', adapter)
        else:
            session = requests
        # the stream=True parameter keeps memory usage low
        r = session.get(
            url, stream=True, allow_redirects=True, headers=HEADERS, timeout=300
        )
        r.raise_for_status()
        try:
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
                        f.flush()
            return local_filename
        except ChunkedEncodingError as err:
            if i == retries:
                raise err
            logging.warning('Download interrupted, retrying...')
            time.sleep(backoff_factor * 2**i)
    raise ValueError("retries must be >= 0")


def http_get(url, etag=None, timeout=600):
    """Download the content from the given URL by making a GET request.

    If an ETag is given, it will do a HEAD request first, to see if the content changed.

    Parameters
    ----------
    url
      The URL to download from.
    etag
      The last ETag to be used for the request (optional).

    Returns
    -------
    A tuple consisting of:
      - The raw content that was downloaded or None if it did not change
      - The new eTag as returned by the HTTP request
    """
    # TODO disable TLS Session IDs and TLS Session Tickets
    #      (plain text cookie visible to anyone who can see the network traffic)
    if etag:
        r = requests.head(url, headers=HEADERS, timeout=timeout)
        r.raise_for_status()
        if 'ETag' in r.headers and etag == r.headers['ETag']:
            return None, etag

    r = requests.get(url, headers=HEADERS, timeout=timeout)
    r.raise_for_status()

    new_etag = None
    if 'ETag' in r.headers:
        new_etag = r.headers['ETag']

    return r.content, new_etag



#!/usr/bin/env python3
#
# install.py - part of the FDroid server tools
# Copyright (C) 2013, Ciaran Gultnieks, ciaran@ciarang.com
# Copyright (C) 2013-2014 Daniel Martí <mvdan@mvdan.cc>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import glob
from argparse import ArgumentParser
import logging

from . import _
from . import common
from .common import SdkToolsPopen
from .exception import FDroidException

options = None
config = None


def devices():
    p = SdkToolsPopen(['adb', "devices"])
    if p.returncode != 0:
        raise FDroidException("An error occured when finding devices: %s" % p.output)
    lines = [line for line in p.output.splitlines() if not line.startswith('* ')]
    if len(lines) < 3:
        return []
    lines = lines[1:-1]
    return [line.split()[0] for line in lines]


def main():

    global options, config

    # Parse command line...
    parser = ArgumentParser(usage="%(prog)s [options] [APPID[:VERCODE] [APPID[:VERCODE] ...]]")
    common.setup_global_opts(parser)
    parser.add_argument("appid", nargs='*', help=_("application ID with optional versionCode in the form APPID[:VERCODE]"))
    parser.add_argument("-a", "--all", action="store_true", default=False,
                        help=_("Install all signed applications available"))
    options = parser.parse_args()

    if not options.appid and not options.all:
        parser.error(_("option %s: If you really want to install all the signed apps, use --all") % "all")

    config = common.read_config(options)

    output_dir = 'repo'
    if not os.path.isdir(output_dir):
        logging.info(_("No signed output directory - nothing to do"))
        sys.exit(0)

    if options.appid:

        vercodes = common.read_pkg_args(options.appid, True)
        common.get_metadata_files(vercodes)     # only check appids
        apks = {appid: None for appid in vercodes}

        # Get the signed APK with the highest vercode
        for apkfile in sorted(glob.glob(os.path.join(output_dir, '*.apk'))):

            try:
                appid, vercode = common.publishednameinfo(apkfile)
            except FDroidException:
                continue
            if appid not in apks:
                continue
            if vercodes[appid] and vercode not in vercodes[appid]:
                continue
            apks[appid] = apkfile

        for appid, apk in apks.items():
            if not apk:
                raise FDroidException(_("No signed APK available for %s") % appid)

    else:

        apks = {common.publishednameinfo(apkfile)[0]: apkfile for apkfile in
                sorted(glob.glob(os.path.join(output_dir, '*.apk')))}

    for appid, apk in apks.items():
        # Get device list each time to avoid device not found errors
        devs = devices()
        if not devs:
            raise FDroidException(_("No attached devices found"))
        logging.info(_("Installing %s...") % apk)
        for dev in devs:
            logging.info(_("Installing '{apkfilename}' on {dev}...").format(apkfilename=apk, dev=dev))
            p = SdkToolsPopen(['adb', "-s", dev, "install", apk])
            fail = ""
            for line in p.output.splitlines():
                if line.startswith("Failure"):
                    fail = line[9:-1]
            if not fail:
                continue

            if fail == "INSTALL_FAILED_ALREADY_EXISTS":
                logging.warning(_('"{apkfilename}" is already installed on {dev}.')
                                .format(apkfilename=apk, dev=dev))
            else:
                raise FDroidException(_("Failed to install '{apkfilename}' on {dev}: {error}")
                                      .format(apkfilename=apk, dev=dev, error=fail))

    logging.info('\n' + _('Finished'))


if __name__ == "__main__":
    main()
